/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/**********************************************************************
 * AcknoledgedDataLink.c
 * target hardware: Pushpin Time Of Flight
 * target compiler: SDCC
 *
 * Last modified 01MAR2004 by lifton & broxton.
 *
 * These functions build on the CommDataLink.c layer of the communication 
 * system to provide reliable communication by means of acknowledged packets.
 * 
 ***********************************************************************/

#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>
#include <DataLink.h>

/*
 * Local function prototypes.
 */
unsigned char getNeighborIndex(unsigned char neighborID);
unsigned char addNeighbor(unsigned char neighborID);
void removeNeighbor(unsigned char neighborID);
void startAckTimer(void);
void stopAckTimer(void);
void generateNeighborhoodID(void);

enum dataLinkStateEnumeration {
  IDLE, AWAITING_ACKNOWLEDGEMENT
};

/*
 * This data structure encapsulates all knowledge about neighboring nodes.
 */
typedef struct NeighborDefinition {
  unsigned char id;             // A 8-bit number unique with respect to this node.
  unsigned char sequenceNumber; // Tag associated with the next packet to send to this neighbor.
  unsigned char requestNumber;  // Tag associated with the next packet to receive from this neighbor.
} Neighbor;

#define MAX_NUM_OF_SYN_RETRIES 10
#define ACK_TIMER_RELOAD 0xB838

// Timer 3 macros.
#define enableTimer3() TMR3CN |= 0x04
#define disableTimer3() TMR3CN &= ~(0x04)
#define enableTimer3Interrupt() EIE2 |= 0x01
#define disableTimer3Interrupt() EIE2 &= ~(0x01)
#define clearTimer3InterruptFlag() TMR3CN &= ~(0x80)
#define timer3FromSystemClock() TMR3CN |= 0x02
#define timer3FromSystemClockDiv12() TMR3CN &= ~(0x02)
#define setTimer3ReloadValue(reload) TMR3RLL = (reload & 0xFF); RMR3RLH = (reload >> 8) & 0xFF
#define setTimer3Value(value) TMR3L = (value & 0xFF); TMR3H = (value >> 8) & 0xFF

Neighbor xdata Neighbors[MAX_NUM_OF_NEIGHBORS];
unsigned char xdata NeighborhoodID;
unsigned char dataLinkState;
unsigned char synRetries;
Packet xdata * currentTransmitPacket;
Packet xdata * receivedSynPacket;

void initializeAcknowledgedDataLink(void) {
  unsigned char i;
  stopAckTimer();
  for (i = 0 ; i < MAX_NUM_OF_NEIGHBORS ; i++) {
    Neighbors[i].id = GLOBALADDRESS;
    Neighbors[i].sequenceNumber = 250;
    Neighbors[i].requestNumber = 250;
  }

  dataLinkState = IDLE;
  synRetries = 0;
  currentTransmitPacket = 0;
  receivedSynPacket = 0;

  /*
   * Normally, we use the first byte of the random seed (code
   * location 0x8000) as our Neighborhood ID.  However, this can
   * be overriden by defining USE_RANDOM_NEIGHBORHOOD_ID.
   */
  #ifdef USE_RANDOM_NEIGHBORHOOD_ID
  generateNeighborhoodID();
  #else
  NeighborhoodID = (unsigned char) (osState.randomSeed);
  #endif
}

/*
 * Assigns NeighbohoodID a one-byte value that is different 
 * from the IDs of all this nodes neighbors and different 
 * from both GLOBALADDRESS and IDEADDRESS.
 */
void generateNeighborhoodID(void) {
  unsigned char index;
  NeighborhoodID = GLOBALADDRESS;
  while (NeighborhoodID == GLOBALADDRESS) {
    NeighborhoodID = (unsigned char) rand();
    for (index=0; index<MAX_NUM_OF_NEIGHBORS; index++) {
      if (NeighborhoodID == Neighbors[index].id || 
          NeighborhoodID == IDEADDRESS || 
          NeighborhoodID == IR_FRAME_START_BYTE) {
        NeighborhoodID = GLOBALADDRESS;
        break;
      }
    }
  }
}

/*
 * This function associates an arbitrary neighbor ID to the
 * index corresponding to that neighbors position in the
 * neighbor list.  Returns MAX_NUMBER_OF_NEIGHBORS if the 
 * neighbor is not listed.
 */
unsigned char getNeighborIndex(unsigned char neighborID) {
  unsigned char i=0;
  while (i < MAX_NUM_OF_NEIGHBORS) {
    if (Neighbors[i].id == neighborID) {
      return i;
    }
    i++;
  }
  return MAX_NUM_OF_NEIGHBORS;
}
/*
 * This function returns the neighborID of the neighbor 
 * stored in location neighborIndex of the Neighbor array.
 * 
 * If no neighbor exists at the requested index, the 
 * GLOBALADDRESS is returned.  
 */
unsigned char getNeighborID(unsigned char neighborIndex) {
  if (neighborIndex < MAX_NUM_OF_NEIGHBORS) {
    return Neighbors[neighborIndex].id;
  } else {
    return GLOBALADDRESS;
  }
}

unsigned char getNumOfNeighbors(void) {
  unsigned char num=0;
  unsigned char index;
  for (index=0; index<MAX_NUM_OF_NEIGHBORS; index++) {
    if (Neighbors[index].id != GLOBALADDRESS) {
      num++;
    }
  }
  return num;
}  

/*
 * This function adds a neighbor to the neighborhood.
 * Returns the index in Neighbors[] where the neighbor
 * was added or MAX_NUM_OF_NEIGHBORS if no space could
 * be allocated.
 */ 
unsigned char addNeighbor(unsigned char neighborID) {
  unsigned char index;
  /* 
   * Check so that we don't add a neighbor twice.
   */
  index = getNeighborIndex(neighborID);
  if (index == MAX_NUM_OF_NEIGHBORS) {
    /* Search for a free slot in our Neighbor table. */
    index = 0;
    while (index < MAX_NUM_OF_NEIGHBORS) {
      if (Neighbors[index].id == GLOBALADDRESS) {
        Neighbors[index].id = neighborID;
        if (NeighborhoodID == neighborID) {
          /*
           * Give up our current ID to this neighbor if they conflict.
           */
          generateNeighborhoodID();
        }
	      return index;
      }
      index++;
    }
    /* No slot available. */
    return MAX_NUM_OF_NEIGHBORS;
  } 
  else {
    return index;
  }
}

/* 
 * Remove a neighbor from the Neighbor list.
 */
void removeNeighbor(unsigned char neighborID) {
  unsigned char i=0;
  while (i < MAX_NUM_OF_NEIGHBORS) {
    if (Neighbors[i].id == neighborID) {
      Neighbors[i].id = GLOBALADDRESS;
      return;
    }
    i++;
  }
}

/* 
 * Calling this function queues a packet up to be sent.  Only one packet at 
 * a time may be queued.  nitiates the process of transmitting a packet.  Only one packet at a 
 * time (the transmitPacket) can be held in the queue.  
 */

unsigned char queuePacket(unsigned char toAddress,	      // Intended recipient's neighborhood ID.
			  unsigned char packetType,	      // Classification of this packet.
			  unsigned int contentSize,	      // The number of bytes of content contained in this packet.
			  unsigned char xdata * contentPtr) { // Location of the content.
  
  unsigned char neighborIndex;
  /*
   * First, check if this neighbor is listed.
   * If the toAddress is GLOBALADDRESS, we will not
   * attempt to add it to the neighbor table.
   */
  if (toAddress != GLOBALADDRESS) {
	if ((neighborIndex = getNeighborIndex(toAddress)) == MAX_NUM_OF_NEIGHBORS) {
      if ((neighborIndex = addNeighbor(toAddress)) == MAX_NUM_OF_NEIGHBORS) {
   	    // Couldn't add this neighbor, so do nothing.
  		return PACKET_NOT_QUEUED;
	  }
	}
  }

  /* 
   * If the dataLinkState is IDLE, we can initialize variables and 
   * pass the packet on to the basic packet queueing mechanism.
   * Otherwise, the comm system is busy and the packet cannot be 
   * queued.
   */
  if (dataLinkState == IDLE) {
    synRetries = 0;
    return queueBasicPacket(toAddress, 
			    packetType, 
			    Neighbors[neighborIndex].sequenceNumber, 
			    contentSize, 
			    contentPtr);
  }
  else {
    return PACKET_NOT_QUEUED;
  }
}

/*
 * This function is called by the CommDataLink.c layer of the 
 * communication system every time a packet is finished being sent.
 */
void packetSentCallback(Packet xdata * packet) {  
  if (packet->packetType == SYN) {
    currentTransmitPacket = packet;
    dataLinkState = AWAITING_ACKNOWLEDGEMENT;
    startAckTimer();
  }
  else if (packet->packetType == ACK) {
    if (receivedSynPacket->contentSize > 0) {
      applicationPacketReceivedCallback(receivedSynPacket);
    }    
  }
  else if (packet->packetType == BROADCAST) {
    applicationPacketSentCallback();
  }
}  


void packetReceivedCallback(Packet xdata *packet) {
  unsigned char neighborIndex;
  /*
   * check if this neighbor is listed.
   */
  if ((neighborIndex = getNeighborIndex(packet->fromAddress)) == MAX_NUM_OF_NEIGHBORS) {
    if ((neighborIndex = addNeighbor(packet->fromAddress)) == MAX_NUM_OF_NEIGHBORS) {
      // Couldn't add this neighbor, so do nothing.
      return;
    }
  }
  
  /*
   * Check which kind of packet is received.
   */
  switch (dataLinkState) {
    case IDLE : {
      switch (packet->packetType) {
        case SYN : {
          if (Neighbors[neighborIndex].requestNumber == packet->sequenceNumber) {
            /*
             * Just received a packet we haven't received before.
             * Increment the request number, acknowledge the packet
             * and then store a pointer to it so it can be processed 
             * _after_ the acknowledgement has been sent.  Note that this 
             * assumes the received packet won't be modified between now and
             * when the ACK packet is finished being sent.
             */
            receivedSynPacket = packet;
            (Neighbors[neighborIndex].requestNumber)++;
            queueBasicPacket(packet->fromAddress,ACK,Neighbors[neighborIndex].requestNumber,0,0);
          }
          else if (0) {
            // meme : reset request number if difference is greater than n.
	        }
          else {
            // meme : ignore?
          }
          break;
        }
        case BROADCAST : {
          /*
           * Broadcast messages do not require acknowledgement.
           */
          if (packet->contentSize > 0) {
            applicationPacketReceivedCallback(packet);
          }
          break;
        }
        case ACK : {
          /*
           * Ideally, this won't happen.  Ignore this case.
           */
          break;
        }
        case NAK : {
          /*
           * Ideally, this won't happen.  Ignore this case.
           */
          break;
        }
        default : {
          break;
        }
      }
    }
    case AWAITING_ACKNOWLEDGEMENT : {
      switch (packet->packetType) {
        case SYN : {
          /*
           * If it's not an ACK, ignore it.
           */
          break;
        }
        case BROADCAST : {
          /*
           * If it's not an ACK, ignore it.
           */
          break;
        }
        case ACK : {
          /*
           * Check if it's the correct ACK.
           */
         if ((packet->fromAddress == currentTransmitPacket->toAddress) && 
              (packet->sequenceNumber == (unsigned char)(Neighbors[neighborIndex].sequenceNumber + 1))) {
            (Neighbors[neighborIndex].sequenceNumber)++;
            dataLinkState = IDLE;
            stopAckTimer();
            applicationPacketSentCallback();
            break;
          }
          break;
        }
        case NAK : {
          /*
           * Need to resend the packet just sent if the intended recipient didn't get it.
           */
          if ((packet->fromAddress == currentTransmitPacket->toAddress) && 
              (packet->sequenceNumber == Neighbors[neighborIndex].sequenceNumber)) {
            stopAckTimer();
            if (++synRetries < MAX_NUM_OF_SYN_RETRIES) {
              if (queueBasicPacket(currentTransmitPacket->toAddress,
                      			       currentTransmitPacket->packetType,
                                   currentTransmitPacket->sequenceNumber,
                                   currentTransmitPacket->contentSize,
                                   currentTransmitPacket->contentPtr) == PACKET_NOT_QUEUED) {
                dataLinkState = IDLE;
                applicationPacketNotSentCallback();
              }
            }
            else {
              dataLinkState = IDLE;
              applicationPacketNotSentCallback();
            }
          }
          break;
        }
        default : {
          break;
        }
      }
    }
    default : {
      break;
    }
  }
}

void headerCRCErrorCallback(void) {
  return;
}

void contentCRCErrorCallback(Packet xdata *packet) {
  unsigned char neighborIndex;
  /*
   * First, check if this neighbor is listed.
   */
  if ((neighborIndex = getNeighborIndex(packet->fromAddress)) == MAX_NUM_OF_NEIGHBORS) {
    if ((neighborIndex = addNeighbor(packet->fromAddress)) == MAX_NUM_OF_NEIGHBORS) {
      // Couldn't add this neighbor, so do nothing.
      return;
    }
  }

  if (dataLinkState == IDLE) {
    queueBasicPacket(packet->fromAddress, 
  		   NAK,
  		   Neighbors[neighborIndex].requestNumber,
  		   0, 
  		   0);
  }
}

void startAckTimer(void) {
  setTimer3Value(ACK_TIMER_RELOAD);
  clearTimer3InterruptFlag();
  enableTimer3Interrupt();
  enableTimer3();
}

void stopAckTimer(void) {
  disableTimer3();
  clearTimer3InterruptFlag();
  disableTimer3Interrupt();
}

/* 
 * The timer 3 interrupt serves as a timeout for unacknowledged packets.
 */
void timer3Interrupt(void) interrupt 14 {
  stopAckTimer();
  if ((dataLinkState == AWAITING_ACKNOWLEDGEMENT) && (++synRetries < MAX_NUM_OF_SYN_RETRIES)) {
    if (queueBasicPacket(currentTransmitPacket->toAddress,
			               currentTransmitPacket->packetType,
                     currentTransmitPacket->sequenceNumber,
                     currentTransmitPacket->contentSize,
                     currentTransmitPacket->contentPtr) == PACKET_NOT_QUEUED) {
      dataLinkState = IDLE;
      applicationPacketNotSentCallback();
    }
  }
  else {
    dataLinkState = IDLE;
    applicationPacketNotSentCallback();
  }
}
