/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*******************************************************************
 * Bertha.h
 * target: PushpinV3
 *
 * Last modified 18Jan2003 by mbroxton.
 *
 * This file defines the memory layout for the Bertha operating 
 * system as implemented on the Cygnal C8051F016 microprocessor.
 *******************************************************************/

#define TRUE 1
#define FALSE 0

/* 
 * Preprocessor directives
 */

/*
 * Enumerations used throughout the operating system.
 */

enum converterSelectOptions {
	RS232SELECT=0, IRSELECT=1
};

enum RGBLEDcolors {
	RGB_RED, RGB_GREEN, RGB_BLUE 
};

enum LEDcolors {
	RED1, RED2, AMBER1, YELLOW1, GREEN1, NOCOLOR
};

enum packetTypes {
	SYN, BROADCAST, ACK, NAK
};

enum queuePacketReturnValues {
  PACKET_NOT_QUEUED, PACKET_QUEUED
};

#define MAX_NUM_OF_NEIGHBORS 20


/*
 * This data structure represents the lowest level at which data
 * can be sent and received over the network.  Note that headerCRC,
 * contentCRC, and contentPtr must all reside at the end of the 
 * definition so that xdataCRC8 can properly calculate the CRC of 
 * the packet.
 */
typedef struct PacketDefinition {
  unsigned char toAddress;	     // Intended recipient's neighborhood ID.
  unsigned char fromAddress;	     // Sender's neighborhood ID.
  unsigned char packetType;	     // Classification of this packet.
  unsigned char sequenceNumber;      // Counter for the number of packets sender has successfully sent.
  unsigned int contentSize;	     // The number of bytes of content contained in this packet.
  unsigned char headerCRC;	     // Cyclic redundancy check sum for the packet header.
  unsigned char contentCRC;          // Cyclic redundancy check sum for the packet content.
  unsigned char xdata * contentPtr;  // A pointer to the content of this packet.
} Packet;


#define IncomingContentBufferSize 100
#define OutgoingContentBufferSize 1

/*
 * The start address of the portion of flash memory used as swap 
 * space for temporarily holding a new version of the OS.
 */
#define swapSpace ((unsigned char code *) 0x3C00) // swapSpace and OS space should be equal = 0x7800/2
#define swapSpaceSize ((unsigned int) 0x3C00)

/*
 * Constants and definitions specific to the microprocessor 
 * hardware.  Either 8051-specific or C8051F016-specific.
 */
#define NUMBEROFADCCHANNELS ((unsigned char) 9) /* Temp sensor + 8 pinned out.       */
#define COMMADCCHANNEL ((unsigned char) 0)      /* ADC channel on comm module.       */
#define IRADCTHRESHOLD ((unsigned int) 0x0200) /* ADC value for which IR detected.  */
#define XdataSize ((unsigned int) 2048)        /* 2Kbytes external memory (RAM).    */
#define FlashSize ((unsigned int) 32896)	      /* Flash memory size.             */
#define FlashPageSize ((unsigned int) 512)	  /* Minimum flash memory erase size.  */
#define MiniPageAddress ((unsigned int) 0x8000) /* Extra flash at top of memory. */
#define MiniPageSize ((unsigned int) 128)	/* Size of extra flash at top of memory.*/
#define ErasedPFragChar ((unsigned char)0xff)  /* Erased flash overwrite character. */
typedef unsigned char DataAddress;        /* 256 bytes.            */
typedef unsigned char IdataAddress;       /* 256 bytes.            */
typedef unsigned int XdataAddress;        /* 2048 bytes.           */
typedef unsigned int CodeAddress;         /* 32 Kbytes.            */
typedef unsigned char ReservedMemoryArea; /* Smallest memory unit. */
typedef union AddressableLong {           /* Utility data type.    */
   unsigned long Long;   /* Access as a single unsigned long.      */
   unsigned int Int[2];  /* Access as individual unsigned integers.*/
   unsigned char Char[4];/* Access as individual unsigned bytes.   */
} AddressableLong;
typedef union AddressableInt {            /* Utility data type.    */
   unsigned int Int;     /* Access as individual unsigned integers.*/
   unsigned char Char[2];/* Access as individual unsigned bytes.   */
} AddressableInt;

/*
 * Timing constants.  Determined in large part by the bandwidth 
 * available for communication, 92160 baud, which translates to 
 * approximately 4600 bytes/second for the 4-way IR communication
 * module.  The largest possible PFrag and associated PFragState 
 * is approximately 2500 bytes.  
 */
#define swapSpaceEraseTime ((unsigned int) 354)  /* milliseconds needed to erase the swapSpace (0x3C00)*/
#define ONESECOND ((unsigned long) 1843318)
#define ONEMILLISECOND ((unsigned long) 1844)

#define GLOBALADDRESS ((unsigned char) 0)
#define IDEADDRESS ((unsigned char) 0xFF)

#define maxNumberOfCandidates ((unsigned char) 5)

typedef struct OSStateDefinition {
	unsigned int randomSeed;
	unsigned int size;
	unsigned int version;
	unsigned char isOld;
	unsigned char crc;
} OSState;

// Flash memory functions.
#define enableFlashErase() PSCTL = 0x03
#define disableFlashErase() PSCTL = 0x00
#define enableFlashWrite() PSCTL = 0x01
#define disableFlashWrite() PSCTL = 0x00

// General interrupt functions.
#define enableGlobalInterrupts() EA = 1
#define disableGlobalInterrupts() EA = 0

// Software-forced reset.
#define reboot() (RSTSRC |= 0x02)
