/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/************************************************************************
 * BerthaCommunication.h
 * target: PushpinV3
 *
 * Josh Lifton
 * lifton@media.mit.edu
 * MIT Media Lab
 * Copyright 2002 all rights reserved.
 *
 * Last modified 21APR2002 by lifton.
 *
 * This file contains all functions needed to handle all levels of 
 * communication.
 * 
 *
 * All communication is packet based and interrupt driven.
 *
 * A packet is composed of a header and content.  A header is 
 * composed of the address of the packet's node of origin, the address
 * of the intended destination, the packet's classification, the
 * number of bytes contained in the packet's content, and the packet's
 * checksum.  
 * 
 * Care must be taken when sending or receiving packets.  In
 * particular, it is important to note that all routines originating
 * from a serial interrupt must terminate before another serial
 * interrupt can be handled.  Thus, it is not possible to send a
 * packet as a direct and immediate response to receiving a packet.
 * To get around this, a packet can be placed in a transmit queue that
 * is dealt with after all other serial interrupts are handled.
 *
 * Once enabled, the communication subsystem's default state should be
 * listening for packets.
 *********************************************************************/

#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>

/*
 * The following are the reload values for 8-bit timer baud rate 
 * generation used by the UART.  Values are calculated assuming 
 * a clock frequency of 22.1184MHz. Note that BAUDRATE must be 
 * greater than a certain minimum value in order for the value of
 * TIMER2IRSERIALRELOAD to be valid.  The resistor and capacitor 
 * values listed in the comments after each reload value indicate 
 * typical values used on the 4-way IR communication module to 
 * generate that baud rate.  Note that in general bit rate and 
 * baud rate are not the same; baud rate is the number of signal 
 * transitions per second whereas bit rate is the number of bits 
 * of information transmitted per second.
 */
#define BAUD9600 0x70
#define BAUD14400 0xA0
#define BAUD19200 0xB9
#define BAUD28800 0xD0
#define BAUD38400 0xDC
#define BAUD57600 0xE8		// 10kOhm, 1800pF
#define BAUD92160 0xF1		// 10kOhm, 1085pF
#define BAUD115200 0xF4		// 10kOhm, 820pF
#define MAXBAUDRATE 0xFE	// 10kOhm, 270pF or 220pF
#define BAUDRATE TH1		// The timer 1 reload register generates baud rate.

/*
 * The following values determine the overall baudrate and the timing needed
 * for the initial data stream used by the recipient to lock onto the IR channel.
 */
#define OVERHEAD 250
#define OVERFLOWSPERBIT 16
#define BITSPERUARTBYTE 10
#define UARTBYTESPERDATABYTE 2
#define CYCLESPERBYTE (UARTBYTESPERDATABYTE*BITSPERUARTBYTE*OVERFLOWSPERBIT*(0xFF-BAUDRATE) + OVERHEAD)
#define NUMBEROFSCANBYTES 3
#define TIMER2TRANSMITRELOAD (0xFFFF - CYCLESPERBYTE)
#define TIMER2RECEIVERELOAD (0xFFFF - NUMBEROFSCANBYTES*CYCLESPERBYTE)
#define RECEIVETIMEOUTRELOAD (0XFFFF - (NUMBEROFSCANBYTES + 1)*CYCLESPERBYTE)
#define BROADCASTADDRESSREPITITIONS 6

/* 
 * These functions are defined as macros for the sake of execution speed and clarity.
 */
#define renewReceiveTimeout() setTimer2Value(RECEIVETIMEOUTRELOAD)

/* 
 * Global state needed throughout the communciation process.  
 * Certain variables, such as IncomingBuffer and OutgoingBuffer,
 * should not be located at 0x0000.  The solution is to place a 
 * variable at 0x0000 that is indifferent to location and is 
 * initialized in a function call (e.g. OSReadyToPropagate is 
 * initialized at startup.  Furthermore, it must be in the same 
 * location throughout all versions of the OS.)
 */
#ifdef SDCC
unsigned char xdata serialState;
#else                 // KEIL Compiler
unsigned char xdata serialState _at_ 0x0000;
#endif

unsigned char xdata commMode = DEFAULTCOMMMODE;
unsigned int xdata currentByte = 0;
unsigned char xdata bytesRemaining = 0;
unsigned char xdata byteBuffer = 0;
unsigned char xdata tempCRC = 0;
Packet xdata receivePacket;
Packet xdata transmitPacket;
unsigned char xdata IncomingBuffer[IncomingBufferSize];
unsigned char xdata OutgoingBuffer[OutgoingBufferSize];
Message xdata * xdata outgoingMessagePtr;

/* 
 * Take the necessary steps to initialize the communication and 
 * messaging system.
 */
void initializeCommunication(void) {
	disableCommunication();
	outgoingMessagePtr = (Message xdata *) OutgoingBuffer;
}

/*
 * This is the final step taken in process to enable reception of packets.
 */
void startScanning(void) {

	serialState = SCANNING;
	if (commMode == FOURWAYIR) {
		enableTimer2();	// Set round-robin channel selection in motion.
	}
	enableSerialReceive();
}
	

/* 
 * Initiates the process of transmitting a packet.  Only one packet at a 
 * time (the transmitPacket) can be held in the queue.  
 */
unsigned char queuePacketForTransmission(PushpinLocalID toAddress, unsigned char packetType, 
                                         XdataAddress contentSize, unsigned char xdata * contentPtr) {
	if (!isPacketInQueue()) {
		transmitPacket.toAddress = toAddress;
		transmitPacket.fromAddress = NeighborhoodID;
		transmitPacket.packetType = packetType;
		transmitPacket.contentSize = contentSize;
		transmitPacket.contentPtr = contentPtr;
		transmitPacket.CRC = packetCRC8(&transmitPacket);
		setPacketInQueue();
		if (attemptToSendPacket()) {
			return PACKETTRANSMITTING;
		}
		return PACKETQUEUED;
	}
	return PACKETNOTQUEUED;
}

/*
 * Starts transmitting a packet if the channel is clear.  Otherwise, 
 * waits a random amount of time and checks again.  If the channel 
 * is still busy, returns without sending the packet.  This process 
 * is only initiated if there is indeed a packet queued to be sent.
 * The communication channel is defined to be busy if data is being 
 * received.
 */
unsigned char attemptToSendPacket(void) {
	/*
	 * Check for a packet to send that is not already being sent.
	 */
	if (isPacketInQueue() && !((serialState==PACKETSENT) || (serialState==SENDINGHEADER) || 
	      (serialState==SENDINGCONTENT) || (serialState==SENDINGTOADDRESS) ||
		  (serialState==PACKETRECEIVED))) {
		/*
		 * Check if channel is open.
		 */
		if (serialState != SCANNING) {
			delay((unsigned int) rand());
			if (serialState != SCANNING) {
				return 0;
			}
		}
		disableSerialReceive();
		clearTransmitInterruptFlag();
		clearReceiveInterruptFlag();
		enableSerialInterrupts();
		switch (commMode) {
			case (RS232IRCONVERTER_RS232) :
				serialState = SENDINGHEADER;
				currentByte = 1;
				SBUF = transmitPacket.toAddress;
				return 1;
				break;
			case (RS232IRCONVERTER_IR) :
			/*
			 * Fall through to next case.
			 */
			case (FOURWAYIR) :
				serialState = SENDINGTOADDRESS;
				currentByte = 1;
				disableTimer2();
				byteBuffer = transmitPacket.toAddress;
				SBUF = encodeNybbleTable[0x0F & byteBuffer];
				byteBuffer >>= 4;
				bytesRemaining = 1;
				return 1;
				break;
			default :
				disableCommunication();
				enableCommunication();
				break;
		}
	}
	return 0;
}

/*
 * This function is called once a complete packet has been sent.  This
 * function manages all the clean-up necessary after sending a packet 
 * over the communication channel.
 */	
void packetSent(void) {
	clearPacketInQueue();
	if (commMode == RS232IRCONVERTER_RS232 || commMode == RS232IRCONVERTER_IR) {
		setCommMode(RS232IRCONVERTER_RS232);
	}
	startScanning();
	nextMessage();
}

/*
 * This function is called once a complete packet has been received.  This 
 * function manages the validation of the received packet and passes the 
 * content on to appropriate handlers.
 */
void packetReceived(void) {

	/*
	 * Don't want to timeout.
	 */
	disableTimer2();

	/*
	 * If building the network neighborhood, just record
	 * the addresses of the packets.
	 */
	if (isBuildingNeighborhood()) {
//		if (receivePacket.CRC == tempCRC) {
//			addNeighbor(receivePacket.fromAddress);
//			addNeighbor(receivePacket.toAddress);
//		}
	}
	
	/*
	 * If the packet is valid, then deal with it according to its type.
	 */
	else if (receivePacket.CRC == packetCRC8(&receivePacket)) {
		processIncomingMessage((Message xdata *) &IncomingBuffer);
	}

	/*
	 * If none of the above, then it's a faulty packet that needs to be 
	 * dealt with according to its type.
	 */
	else {
		packetReceiveError();
	}

	/*
	 * Now that this packet is dealt with, start the communication 
	 * subsystem listening for other packets.
	 */
	startScanning();

	/*
	 * Finally, send any packets that may have been queued.
	 */
	attemptToSendPacket();
}

void packetReceiveError(void) {
	disableCommunication();
	enableCommunication();
	return;
}

/*
 * Determines the hardware used for the communication module.
 * Currently, only the four-way IR module and the RS232-to-IR 
 * module are supported.  Aside from detecting which module is 
 * present, this function also configures the processing module 
 * hardware for use with the detected module.  Finally, the 
 * software is configured for use with the detected module by 
 * calling setCommMode.
 */
void detectCommModule(void) {
	setUARTMode1();			// 8-bit serial communication.
	T2CON = 0x01;			// Configure timer 2 (no auto-reload).
	CKCON |= 0x20;			// Set timer 2 to run off system clock.
	TMOD |= 0x20;			// Set timer1 as 8-bit auto reload.
	PCON |= 0x80;			// Double baud rate; ie, SMOD = 1.

	/*
	 * Determine the mode of communication.  The RS232<->IR module 
	 * should have a reading of nearly 0 on the ADC.
	 */
	if (getADCValue(COMMADCCHANNEL) > IRADCTHRESHOLD) {
		commMode = FOURWAYIR;
		PRT1CF &= 0x0F;			//|
		IRChannel0 = 1;			//| 
		IRChannel1 = 1;			//| Set pins P1.[4-7] as inputs.
		IRChannel2 = 1;			//|
		IRChannel3 = 1;			//|
	}
	else {
		/*
		 * The RS232-to-IR module should be in RS232 mode as a default.
		 * Any incoming IR traffic will generate an external interrupt 
		 * that then causes the module to switch to IR mode when needed.
		 */
		commMode = RS232IRCONVERTER_RS232;
		PRT1CF &= ~(0x20); 		// Set the IR external interrupt (P1.5) as an input.
		PRT1CF |= 0x10;			// Set converterSelect (P1.4) as an output.
	}

	/*
	 * Set the commMode based on the hardware.
	 */
	setCommMode(commMode);
}


void setCommMode(unsigned char mode) {
	/*
	 * Check the communication hardware to determine allowable options.
	 * The previous value of commMode indicates the hardware.
	 */
	switch (commMode) {
		case RS232IRCONVERTER_RS232 :
			/*
			 * Fall through to next case.
			 */
		case RS232IRCONVERTER_IR :
			if (mode == RS232IRCONVERTER_IR) {
				BAUDRATE = BAUD92160;
				disableExternal5Interrupt();
				clearExternal5InterruptFlag();
				converterSelect = IRSELECT;
			}
			else if (mode == RS232IRCONVERTER_RS232) {
				BAUDRATE = BAUD19200;	// Set timer1 reload for 19200 bits per second.
				clearExternal5InterruptFlag();
				enableExternal5Interrupt();
				converterSelect = RS232SELECT;
			}
			break;
		case FOURWAYIR :
			BAUDRATE = BAUD92160;
			break;
		default :
			break;
	}

	/*
	 * Set commMode to the specified value.
	 */
	commMode = mode;
}

/*
 * Initiates an interrupt-driven scan and reception of a packet.
 */
void enableCommunication(void) {
	unsigned char i;
	for (i=0; i<IncomingBufferSize; i++) {
		IncomingBuffer[i] = 0x11;
	}
	for (i=0; i<OutgoingBufferSize; i++) {
		OutgoingBuffer[i] = 0x22;
	}
	currentByte = 0;
	bytesRemaining = 0;
	enableTimer1();
	clearTimer2InterruptFlag();
	enableTimer2Interrupt();
	clearTransmitInterruptFlag();
	clearReceiveInterruptFlag();
	clearPacketInQueue();
	enableSerialInterrupts();
	switch (commMode) {
		case RS232IRCONVERTER_RS232 :
		/*
		 * Fall through to next case.  Always reset into RS232 mode 
		 * when using the RS232-to-IR converter module.
		 */
		case RS232IRCONVERTER_IR :
			setCommMode(RS232IRCONVERTER_RS232);
			break;
		case FOURWAYIR :
			setCommMode(FOURWAYIR);
			break;
		default :
			setCommMode(commMode);
			break;
	}
	startScanning();
}

/* 
 * Shuts down communication.
 */
void disableCommunication(void) {
	serialState = DISABLED;
	disableTimer1();
	disableTimer2();
	disableTimer2Interrupt();
	disableSerialInterrupts();
	disableSerialReceive();
	clearTimer2InterruptFlag();
	clearTransmitInterruptFlag();
	clearReceiveInterruptFlag();
	clearPacketInQueue();
//	clearNWBusy();
}

/*
 * Handles received bytes by building up a packet one byte at a time.  This function 
 * is independent of the physical layer of communication used.
 */
void setNextPacketByte(unsigned char c) {
	switch (serialState) {
		case SCANNING :
			if (c == NeighborhoodID || c == GLOBALADDRESS || isBuildingNeighborhood()) {  
				//listenToAllBytes();
				receivePacket.toAddress = 0;
				receivePacket.fromAddress = 0;
				receivePacket.packetType = 0;
				receivePacket.contentSize = 0;
				receivePacket.CRC = 0;
				currentByte = 0;
				serialState = RECEIVINGHEADER;	// Fall through to next case.
			}
			else {
				/*
				 * Ignore all all bytes not addressed to this Pushpin.
				 * This includes those bytes from incomplete packets.
				 */
				break;
			}
		case RECEIVINGHEADER :
			if ((c == receivePacket.toAddress)  && (currentByte == 1)) {		// Wait until end of toAddress broadcast.
				break;
			}
			*(((unsigned char xdata *) &receivePacket) + currentByte++) = c;
			if (currentByte >= PACKETHEADERSIZE) {	// Finished receiving the header.  Decide what to do with content.
				if (isBuildingNeighborhood()) {
					currentByte = 0;
					tempCRC = 0;
					tempCRC = CRC8Table[tempCRC ^ receivePacket.toAddress];
					tempCRC = CRC8Table[tempCRC ^ receivePacket.fromAddress];
					tempCRC = CRC8Table[tempCRC ^ receivePacket.packetType];
					tempCRC = CRC8Table[tempCRC ^ *((unsigned char xdata *) &(receivePacket.contentSize))];
					tempCRC = CRC8Table[tempCRC ^ *(((unsigned char xdata *) &(receivePacket.contentSize)) + 1)];
					if (receivePacket.contentSize) {
						serialState = IGNORINGCONTENT;
					}
					else {
						serialState = PACKETRECEIVED;
					}
				}
				else if (receivePacket.contentSize != 0) {
					currentByte = 0;
					if (setContentPtr(&receivePacket)) {
						serialState = RECEIVINGCONTENT;
					}
					else {
						serialState = IGNORINGCONTENT;
					}
				}
				else {
					serialState = PACKETRECEIVED;
				}
			}
			break;
		case RECEIVINGCONTENT :
			if (receivePacket.packetType) {
				/*
				 * Received bytes should be stored in extended RAM.
				 */
				*(receivePacket.contentPtr + currentByte++) = c;
			}
			else {
				/*
				 * Received bytes should be stored in flash.
				 */
				receivePacket.contentPtr += currentByte;
				disableGlobalInterrupts();  // Must disable interrupts to avoid extra MOVX executions.
				enableFlashWrite();         // The next MOVX instruction will effect flash.
				*(receivePacket.contentPtr) = c;
				disableFlashWrite();
				enableGlobalInterrupts();
				receivePacket.contentPtr -= currentByte;
				currentByte++;
			}
			if (currentByte >= receivePacket.contentSize) {
				serialState = PACKETRECEIVED;
			}
			break;
		case IGNORINGCONTENT :
			/*
			 * Don't want to tie up the receive buffer by writing ignored data to it.
			 */
			currentByte++;
			if (isBuildingNeighborhood()) {
				tempCRC = CRC8Table[tempCRC ^ c];
			}
			if (currentByte >= receivePacket.contentSize) {
				serialState = PACKETRECEIVED;
			}
			break;
		default :
			break;
	}
}

/*
 * Returns the next byte to send out over the communication link.
 */
unsigned char getNextPacketByte() {
	unsigned char xdata next;
	switch (serialState) {
		case SENDINGHEADER :
			next = *(((unsigned char *) &transmitPacket) + (currentByte++));
			if (currentByte >= PACKETHEADERSIZE) {
				currentByte = 0;
				if (transmitPacket.contentSize > 0) {
					serialState = SENDINGCONTENT;
				}
				else {
					serialState = PACKETSENT;
				}
			}				
			break;
		case SENDINGCONTENT :
			if (transmitPacket.packetType) {
				/*
				 * Bytes to be sent are stored in extended RAM.
				 */
				next = *(transmitPacket.contentPtr + (currentByte++));
			}
			else {
				/*
				 * Bytes to be sent are stored in flash.
				 */
				next = *(((unsigned char code *)(transmitPacket.contentPtr)) + (currentByte++));
			}
			if (currentByte >= transmitPacket.contentSize) {
				serialState = PACKETSENT;			
			}
			break;
		default :
			break;
	}
	return next;
}

/*
 * Determines where to put the data contained in the incoming packet based on 
 * information contained in the header.  Returns a null pointer if the request 
 * cannot be accommodated for any reason (space limitations, invalid header, etc.).
 */
unsigned char setContentPtr(Packet xdata * packet) {
	if (serialState == RECEIVINGHEADER) {
		if (receivePacket.packetType) {
			if (receivePacket.contentSize <= IncomingBufferSize) {
			/*
			 * meme : Should lock IncomingBuffer so nothing else accesses it.
			 */
				packet->contentPtr = IncomingBuffer;
				return 1;
			}
		}
		else {
			/*
			 * meme : For now, there is simply a chunk of flash defined as swap space.
			 */
			if (receivePacket.contentSize <= swapSpaceSize) {
				packet->contentPtr = (unsigned char xdata *) swapSpace;
				return 1;
			}
		}
	}
	return 0;
}


/*
 * Note on the UART:
 * Interrupt number 4 is the serial interrupt.  A serial interrupt is called any time
 * RI or TI is set to 1.  This can be done in either software or hardware.  RI is set 
 * to 1 by the UART hardware whenever the UART has completed receiving a byte.  The 
 * byte just received resides in SBUF.  TI is set to 1 by the UART hardware whenever 
 * the UART has completed transmitting a byte.  Loading a byte into SBUF initiates 
 * the transmission of that byte by the UART.  Thus, the statement SBUF=SBUF instructs 
 * the UART to transmit the byte it last received.  Both TI and RI are only cleared to 
 * 0 in software; the hardware does not clear them at all.
 * 
 * This interrupt service routine checks the state of the transmission or 
 * reception processes and acts accordingly; it either transmits the next byte 
 * or it logs the byte just received and waits for the next byte.  The byteHalfSent 
 * variable keeps tabs on which half of the data byte (high nybble or low nybble) 
 * is being dealt with.  All other global state regarding serial communication is
 * held in the serialState variable.
 */
void serialInterrupt (void) interrupt 4 {
	/* 
	 * The receive interrupt (RI) indicates that the UART has completed 
	 * receiving a byte.  
	 */
	if (RI) {
		clearReceiveInterruptFlag();
		renewReceiveTimeout();
		enableTimer2();
		switch (commMode) {
			case RS232IRCONVERTER_RS232 :
			/*
			 * Each network byte is a data byte.
			 */
				setNextPacketByte(SBUF);
				break;
			case RS232IRCONVERTER_IR :
			/*
			 * Fall through to next case.
			 */
			case FOURWAYIR :	
			/* 
			 * Two network bytes for each data byte.
			 */
				if (!bytesRemaining) {
					byteBuffer = decodeNybbleTable[SBUF];
					bytesRemaining = 1;
				}
				else {
					byteBuffer |= (decodeNybbleTable[SBUF] << 4);
					bytesRemaining = 0;
					setNextPacketByte(byteBuffer);
				}
				break;
			default :			
			/*
			 * This should never happen.  Something has gone wrong.
			 */
				break;
		}
		if (serialState == PACKETRECEIVED) {
			packetReceived();
		}
	}
	if (TI) {
		clearTransmitInterruptFlag();
		switch (commMode) {
			case RS232IRCONVERTER_RS232 :
				if (serialState != PACKETSENT) {
					SBUF = getNextPacketByte();
				}
				else {
					packetSent();
				}
				break;
			case RS232IRCONVERTER_IR :
			/*
			 * Fall through to next case.
			 */
			case FOURWAYIR :
				if (bytesRemaining) {
					/*
					 * Send second half of the byte being sent.
					 */
					SBUF = encodeNybbleTable[0x0F & byteBuffer];
					bytesRemaining = 0;
				}
				else if (serialState == SENDINGTOADDRESS) {
					/*
					 * Start countdown to sending next byte.
					 */
				        setTimer2Value(TIMER2TRANSMITRELOAD);
					clearTimer2InterruptFlag();
					enableTimer2();
				}
				else if (serialState != PACKETSENT) {
					/*
					 * Send the next byte in the packet.
					 */
					byteBuffer = getNextPacketByte();
					SBUF = encodeNybbleTable[0x0F & byteBuffer];
					byteBuffer >>= 4; 
					bytesRemaining = 1;
				}
				else if (serialState == PACKETSENT) {
					packetSent();
				}
				break;
			default :	
			/* 
			 * Should never get here.  Something is wrong.
			 */
			  break;
		}
	}
}

void timer2Interrupt(void) interrupt 5 {
	/* 
	 * Nothing interesting on current channel, so switch to the next channel.
	 */
	if (commMode == FOURWAYIR && serialState == SCANNING) {
		IRChannelSelect0 = ~IRChannelSelect0;
		if (!IRChannelSelect0) {
			IRChannelSelect1 = ~IRChannelSelect1;
		}
		setTimer2Value(TIMER2RECEIVERELOAD);
	}

	/*
	 * Wait a while before sending next address.
	 */
	else if ((commMode == FOURWAYIR || commMode == RS232IRCONVERTER_IR) && serialState == SENDINGTOADDRESS) {
		if (currentByte++ < BROADCASTADDRESSREPITITIONS) {
			byteBuffer = transmitPacket.toAddress;
			SBUF = encodeNybbleTable[0x0F & byteBuffer];
			byteBuffer >>= 4; 
			bytesRemaining = 1;
		}
		else {
			serialState = SENDINGHEADER;
			currentByte = 1;
			byteBuffer = getNextPacketByte();
			SBUF = encodeNybbleTable[0x0F & byteBuffer];
			byteBuffer >>= 4; 
			bytesRemaining = 1;
		}
		disableTimer2();
	}

	/*
	 * The rest of the packet has gone missing.  Reset communication state.
	 */
	else if (REN) {
		disableCommunication();
		enableCommunication();
	}

	/*
	 * Clear the interrupt flag in any case.
	 */
	clearTimer2InterruptFlag();
}

/*
 * This interrupt is triggered whenever a falling edge on P1.5 is detected and 
 * EX5=1 and EA=1 (this interrupt is enabled and globel interrupts are enabled).
 * Currently, the only time this is used is when the RS232-to-IR converter 
 * communication module is in place, in which case this interrupt signals that 
 * IR communication is being received and causes the module to switch to IR 
 * mode.

	meme : should disable this interrupt when transmitting over IR.

 */
void external5Interrupt(void) interrupt 17 {
	if (commMode == RS232IRCONVERTER_RS232 || commMode == RS232IRCONVERTER_IR) {
		disableExternal5Interrupt();		// This interrupt is not needed for a while.
		setCommMode(RS232IRCONVERTER_IR);	// Start listening over IR.
	}
	clearExternal5InterruptFlag();
}

/* 
 * Adapted from: http://cell-relay.indiana.edu/mhonarc/cell-relay/1999-Jan/msg00074.html
 * 8 bit CRC Generator, MSB shifted first
 * Polynom: x^8 + x^2 + x^1 + 1
 * 
 * Calculates an 8-bit cyclic redundancy check sum for a packet.
 * This function takes care not to include the packet's check sum in calculating 
 * the check sum.  Assumes the CRC is the last byte of the packet header.  Also
 * takes care to look in code space (instead of xdata space) when dealing with a 
 * PFrag.
 */
unsigned char packetCRC8(Packet xdata * packet) {
	unsigned int xdata i;
	unsigned char xdata crc = 0;
	for (i=0; i < (PACKETHEADERSIZE - 1); i++) {	// Don't include the CRC itself when calculating the CRC.
		crc = CRC8Table[crc ^ (*(((unsigned char xdata *) packet) + i))];
	}
	switch ((*packet).packetType) {
		case DIRECTTOFLASH :
			for (i=0; i < (*packet).contentSize; i++) {
				crc = CRC8Table[crc ^ (*((unsigned char code *) (packet->contentPtr + i)))];
			}
			break;
		default :
			for (i=0; i < (*packet).contentSize; i++) {
				crc = CRC8Table[crc ^ (*(((*packet).contentPtr) + i))];
			}
			break;
	}
	return crc;
}
