/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/***************************************************************************
 * BerthaFunctions.c
 * target: PushpinV3
 *
 *
 * Last modified 07MAR2002 by lifton.
 *
 * These functions are for Bertha's internal use only.
 *************************************************************************/

#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>


// ADC functions.
#define getADCChannel() (AMX0SL)
#define setADCChannel(channel) (AMX0SL = channel)
#define clearADCWindowInterruptFlag() (ADWINT = 0)
#define enableADCWindowInterrupt() (EIE1 |= 0x04)
#define disableADCWindowInterrupt() (EIE1 &= ~(0x04))
#define setADCWindowGreaterThanValue(value) ADC0GTL = (value & 0xFF); ADC0GTH = (value >> 8) & 0xFF
#define setADCWindowLessThanValue(value) ADC0LTL = (value & 0xFF); ADC0LTH = (value >> 8) & 0xFF


/*
 * Configure and enable analog to digital converter.  Select channel 0 as default.
 */
//void enableADC(void) {
//	REF0CN |= 0x03;		// Enable VREF and ADC/DAC bias.
//	AMX0CF = 0x00;		// Single-ended inputs (not differential).
//	AMX0SL = 0x00;		// Select Analog Input 0 (AIN0).
//	ADC0CF &= ~(0x07);	// Gain 1.
//	ADC0CF |= 0xE0;		// Conversion clock to system clock / 16.
//	ADLJST = 0;			// Right justification.
//	ADCTM = 1;			// Tracking is software controlled
//						// necessary for software control through ADBUSY 		
//	ADCEN = 1;			// Enable ADC.
//}

/*
 * Returns the current 10-bit value read from the specified ADC channel.  Assumes 
 * the data format of the ADC is right justified.  There are two issues to keep 
 * in mind when using the ADC.  First, there is some settling time required 
 * after changing channels.  If the impedance coming into the ADC channel is high, 
 * the RC time constant will not be negligible.  In particular, the voltage 
 * dividers used to identify the various communication and expansion modules are
 * typically high impedance.  Thus, some amount of delay may be required between 
 * switching ADC channels and detecting the communication or expansion modules.
 * The other issue is perhaps more critical.  It appears that the first ADC 
 * reading taken after changing channels comes from the previous channel.  Why 
 * this is so is not known, although one possibility is that the analog mux 
 * controlling access to the ADC doesn't actually change channels until a reading
 * is taken.  That is, changing the AMX0SL doesn't switch the mux in itself, 
 * but only lets the ADC know which channel to switch to when a reading is 
 * started (ADBUSY = 1).  Thus, that the first reading appears to be from the 
 * previous channel is just a result of the ADC being much faster than the RC 
 * time constant associated with the input.
 * Update: see Cygnal application note.
 */
unsigned int getADCValue(unsigned char channel) {
	AddressableInt ADCvalue;
	if (enableADCChannel(channel)) {
		ADBUSY = 1;				// Start ADC tracking and conversion.
		while (!ADCINT);		// Wait for conversion to end.
		ADCINT = 0;				// Clear interrupt.
		ADCvalue.Char[0] = ADC0H;
		ADCvalue.Char[1] = ADC0L;
		return ADCvalue.Int;
	}
	return 0;
}

/*
 * Returns '1' if the requested channel has been switched to and the ADC 
 * is ready to sample.  Returns '0' if the channel requested doesn't exist.
 */
unsigned char enableADCChannel(unsigned char channel) {
	if (channel < NUMBEROFADCCHANNELS) {
		REF0CN |= 0x03;		// Enable VREF and ADC/DAC bias.
		AMX0CF = 0x00;		// Single-ended inputs (not differential).
		ADC0CF &= ~(0x07);	// Gain 1.
		ADC0CF |= 0xE0;		// Conversion clock to system clock / 16.
		ADLJST = 0;			// Right justification.
		ADCTM = 1;			// Tracking is software controlled
							// necessary for software control through ADBUSY 		
		ADCEN = 1;			// Enable ADC.
		if (AMX0SL != channel) {
			AMX0SL = channel;	// Set analog mux channel select.
			ADBUSY = 1;			//| Take one reading so as to force the mux to 
			ADCINT = 0;			//|
			while (!ADCINT);	//| switch and let input settle according to 
			ADCINT = 0;			//| its RC time constant.
		}
		return 1;
	}
	return 0;
}

void configurePins(void) {
	PRT0CF = 0xFF;	// Set all of port 0 to push-pull outputs.
	PRT1CF = 0xFF;	// Set all of port 1 to push-pull outputs.
	PRT2CF = 0xFF;	// Set all of port 2 (not pinned out) to push-pull outputs.
	XBR0 |= 0x04;	// UART TX and RX on P0.0 and P0.1.	
	XBR2 |= 0x40;	// Enable crossbar. 
	XBR2 |= 0x80;	// Disable weak pull-ups.
}

void disableWatchdogTimer(void) {
	WDTCN = 0xde;	// This register must be set first.
	WDTCN = 0xad;	// This register must be set second.
}

void enableExternalClock(void) {
	CKCON |= 0x10;				// Don't divide by 12 for base clock; ie, T1M=1.
	OSCXCN = 0x67;				// Start external oscillator > 6.74MHz (using a 22.118MHz crystal)
	delay(0xffff);				// Wait more than a millisecond.
	while (!(OSCXCN & 0x80)) {} // Wait until external oscillator is stable.
	OSCICN |= 0x80;				// Check for missing clock.
	OSCICN |= 0x08;				// Use external crystal.
	FLSCL = ((FLSCL & 0xF0) | 0x09);	// Set flash scalar for 22Mhz clock.
}

void initializeHardware(void) {
	disableWatchdogTimer();
	enableExternalClock();
	configurePins();
	statusLED = 0;
}

/*
 * Get a random number from the ADC Inputs 
 */
unsigned int generateRandomNumberFromADC(void) {
  unsigned int i,j;
  unsigned int ADCSum, randomNumber;
  unsigned char randomBit;

  for (j = 0; j < (sizeof(unsigned int) * 8); j++) {
    ADCSum = 0;

    // Find the sum of the samples from each ADC channel
    for (i = 0; i < NUMBEROFADCCHANNELS; i++) {
      ADCSum += getADCValue(i);
    }

    // XOR the bits in ADCSum into a single random bit
    for (i = 0; i < (sizeof(unsigned int) * 8); i++) {
      randomBit ^= (ADCSum >> i) & 0x01;
    }

    // Make this bit the next bit in our random number
    randomNumber = (randomNumber << 1) | randomBit;
  }

  return randomNumber; 
}
					    
/*
 * A wrapper for the srand() function provided in rand.h.
 */
void generateRandomSeed(void) {
  srand(generateRandomNumberFromADC());
}

/*
 * A blocking function that flashes the LED located on the processing
 * module a certain number of times at a certain interval, both given 
 * as arguments. To be replaced by use of a timer.
 */
void flashLED(unsigned int times, unsigned int interval) {
	while(times--) {
		statusLED = 1;
		delay(interval);
		statusLED = 0;
		delay(interval);
	}
}

/* 
 * A blocking function that simply waits an amount of time roughly 
 * proportional to the value of the argument.
 */
void delay(unsigned int interval) {
	while(interval--);
}
