/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/**************************************************************************
 * BerthaTiming.c
 * target: PushpinV3
 *
 * Last modified 08APR2002 by lifton.
 *
 * Implementation of Bertha's timing functions, including the real-time clock.
 * Assumes the Pushpin is running off a 22.1184MHz crystal.  A 32-bit real-time
 * clock is formed by conjoining timer 0 running in 16-bit clock mode with a 
 * 16-bit unsigned integer incremented every time timer 0 overflows.  Timer 0 
 * is configured increment once for every 12 system clocks.  Given this 
 * arrangement, the 32-bit real-time clock should overflow once every 38.84 
 * minutes.  The observed overflow rate, however, is approximately twice this 
 * amount, perhaps due to how system cycles are counted.
 *
 * Software alarms can be set such that an interrupt is generated after a 
 * specified amount of time, up to the limit of the real-time clock.  Timer 3  
 * triggers the actual alarm.  Overhead between calling the setAlarm function 
 * and actually setting the alarm accounts for approximately 32 microseconds.  
 * This overhead time is taken into account when setting the alarm.
 *
 *************************************************************************/

#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>

/*
 * Macros for controlling timer 0.
 */
#define enableTimer0() TR0 = 1
#define disableTimer0() TR0 = 0
#define enableTimer0Interrupt() ET0 = 1
#define disableTimer0Interrupt() ET0 = 0
#define clearTimer0InterruptFlag() TF0 = 0
#define timer0FromSystemClock() CKCON |= 0x08
#define timer0FromSystemClockDiv12() CKCON &= ~(0x08)

unsigned int data highTime;

void startRealTimeClock(void) {
    /*
     * Timer 0 is the lower 16 bits of the 32 bit real-time clock.
     */
    disableTimer0();
    enableTimer0Interrupt();
    clearTimer0InterruptFlag();
	timer0FromSystemClockDiv12();
    TMOD &= ~(0x0F);    // Clear timer 0 control registers.
    TMOD |= 0x01;       // Set timer 0 as 16-bit timer.
	highTime = 0;
    setTime(0);
    enableTimer0();
}

unsigned long getTime(void) {
  AddressableLong data time;
  disableTimer0();
  /* 
   * If getTime is called from an interrupt routine, the timer may
   * have overflowed while that interrupt routine was being run, hence 
   * highTime has not yet been incremented.  To correct for this, we
   * must check the interrupt pending flag and add 1 to highTime
   * manually if an interrupt has occured that has not yet been
   * processed.
   */
  if (TF0) {
    time.Int[1] = highTime + 1;
  } else {
    time.Int[1] = highTime;
  }
  time.Char[1] = TH0;
  time.Char[0] = TL0;
  enableTimer0();
  return time.Long;
}

void setTime(unsigned long time) {
    TL0 = *(((unsigned char *) &time) + 3);
    TH0 = *(((unsigned char *) &time) + 2);
	highTime = *((unsigned int *) &time);
}

void timer0Interrupt(void) interrupt 1 {
    clearTimer0InterruptFlag();
	highTime++;
}
