/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*******************************************************************************
 * CommDataLink.c
 * target hardware: Pushpin Time Of Flight
 * target compiler: SDCC
 *
 * Last modified 24FEB2004 by lifton.
 *
 * This file contains all functions needed to handle the lowest layer of the 
 * Pushpin networking stack.
 * 
 *******************************************************************************/

#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>
#include <DataLink.h>

/*
 * These constants define the size of the packet header for different circumstances.
 */
#define HEADER_SIZE_WITHOUT_CRCS (sizeof(Packet) - 4)  // headerCRC + contentCRC + contentPtr = 4 bytes
#define HEADER_SIZE_WITH_CRCS (sizeof(Packet) - 2)     // contentPtr = 2 bytes

/*
 * The following are the reload values for 8-bit timer baud rate 
 * generation used by the UART.  Values are calculated assuming 
 * a clock frequency of 22.1184MHz. Note that BAUDRATE must be 
 * greater than a certain minimum value in order for the value of
 * TIMER2IRSERIALRELOAD to be valid.  The resistor and capacitor 
 * values listed in the comments after each reload value indicate 
 * typical values used on the 4-way IR communication module to 
 * generate that baud rate.  Note that in general bit rate and 
 * baud rate are not the same; baud rate is the number of signal 
 * transitions per second whereas bit rate is the number of bits 
 * of information transmitted per second.
 */
#define BAUD9600 0x70
#define BAUD14400 0xA0
#define BAUD19200 0xB9
#define BAUD28800 0xD0
#define BAUD38400 0xDC
#define BAUD57600 0xE8		// 10kOhm, 1800pF
#define BAUD92160 0xF1		// 10kOhm, 1085pF
#define BAUD115200 0xF4		// 10kOhm, 820pF
#define MAXBAUDRATE 0xFE	// 10kOhm, 270pF or 220pF
#define BAUDRATE TH1		// The timer 1 reload register generates baud rate.

/*
 * These definitions determine the length and structure of the preamble
 * used to help the receiving node lock onto the correct IR channel.
 */
#define OVERHEAD 250
#define OVERFLOWSPERBIT 16
#define BITSPERUARTBYTE 10
#define UARTBYTESPERDATABYTE 2
#define CYCLESPERBYTE (UARTBYTESPERDATABYTE*BITSPERUARTBYTE*OVERFLOWSPERBIT*(0xFF-BAUDRATE) + OVERHEAD)
#define NUMBEROFSCANBYTES 3
#define TIMER2TRANSMITRELOAD (0xFFFF - CYCLESPERBYTE)
#define TIMER2RECEIVERELOAD (0xFFFF - NUMBEROFSCANBYTES*CYCLESPERBYTE)
#define RECEIVETIMEOUTRELOAD (0XFFFF - (NUMBEROFSCANBYTES + 1)*CYCLESPERBYTE)
#define BROADCASTADDRESSREPITITIONS 8  // meme : this should work at 6... may still need to be adjusted


enum serialStatesEnumeration {
  DISABLED, IDLE, SCANNING, SENDINGPREAMBLE, RECEIVINGHEADER, 
  RECEIVINGCONTENT, IGNORINGCONTENT, SENDINGHEADER, SENDINGCONTENT, 
  PACKETRECEIVED, PACKETSENT, STASIS, LOCKED_ON_FRAME
};

/*
 * Function prototypes and argument descriptions for locally used functions.
 */
void startScanning(void);
void packetSent(void);
unsigned char xdataCRC8(unsigned char xdata * bufferPtr,     // A pointer to the start of the data buffer to check.
			unsigned int bufferSize);            // Number of bytes in the data buffer to check.
void setNextReceivePacketByte(unsigned char c);              // Byte to be added to the receive packet.
unsigned char getNextTransmitPacketByte(void);
unsigned char setContentPtr(Packet xdata * packet);          // Pointer to the packet in which to set the content pointer.

/*
 * State variables.
 */
unsigned char xdata serialState;
unsigned char xdata sequenceNumber;
unsigned char isPacketInQueue;  // meme : used to be a bit, probably should be still.
unsigned int xdata currentByte;
unsigned char xdata bytesRemaining;
unsigned char xdata byteBuffer;

/*
 * Buffers for storing incoming and outgoing content and headers.
 */
unsigned char xdata IncomingContentBuffer[IncomingContentBufferSize];
unsigned char xdata OutgoingContentBuffer[OutgoingContentBufferSize];
Packet xdata receivePacket;
Packet xdata transmitPacket;

/* 
 * These functions are defined as macros for the sake of execution speed and clarity.
 */
#define renewReceiveTimeout() setTimer2Value(RECEIVETIMEOUTRELOAD)

/*
 * Macros for controlling the serial UART.
 */
#define enableSerialInterrupts() ES = 1
#define disableSerialInterrupts() ES = 0
#define enableSerialReceive() REN = 1
#define disableSerialReceive() REN = 0
#define setTransmitInterruptFlag() TI = 1
#define clearTransmitInterruptFlag() TI = 0
#define clearReceiveInterruptFlag() RI = 0
#define setUARTMode1() SM0 = 0; SM1 = 1


/*
 * Macros for controlling timer 1.
 */
#define enableTimer1() TR1 = 1
#define disableTimer1() TR1 = 0
#define enableTimer1Interrupt() ET1 = 1
#define disableTimer1Interrupt() ET1 = 0
#define clearTimer1InterruptFlag() TF1 = 0

/*
 * Macros for controlling timer 2.
 */
#define enableTimer2() TR2 = 1
#define disableTimer2() TR2 = 0
#define enableTimer2Interrupt() ET2 = 1
#define disableTimer2Interrupt() ET2 = 0
#define clearTimer2InterruptFlag() TF2 = 0
#define setTimer2ReloadValue(reload) RCAP2L = (reload & 0xFF); RCAP2H = (reload >> 8) & 0xFF
#define setTimer2Value(value) TL2 = (value & 0xFF); TH2 = (value >> 8) & 0xFF


/******************************************************************************************************FUNCTIONS*********************/

/*
 * This should be the last function called in the process of preparing to receive IR communications.
 */
void startScanning(void) {
  serialState = SCANNING;
  clearTimer2InterruptFlag();
  setTimer2Value(TIMER2RECEIVERELOAD);
  enableTimer2();	// Set round-robin channel selection in motion.
  enableSerialReceive();
}

/* 
 * Calling this function queues a packet up to be sent.  Only one packet at 
 * a time may be queued.  nitiates the process of transmitting a packet.  Only one packet at a 
 * time (the transmitPacket) can be held in the queue.  
 */
unsigned char queueBasicPacket(unsigned char toAddress,	      // Intended recipient's neighborhood ID.
			  unsigned char packetType,	      // Classification of this packet.
			  unsigned char sequenceNumber,       // Counter for the number of packets sender has successfully sent.
			  unsigned int contentSize,	      // The number of bytes of content contained in this packet.
			  unsigned char xdata * contentPtr) { // Location of the content.
  if (!isPacketInQueue) {
    isPacketInQueue = TRUE;
    transmitPacket.toAddress = toAddress;
    transmitPacket.fromAddress = NeighborhoodID;
    transmitPacket.packetType = packetType;
    transmitPacket.sequenceNumber = sequenceNumber;
    transmitPacket.contentSize = contentSize;
    transmitPacket.contentPtr = contentPtr;
    transmitPacket.headerCRC = xdataCRC8((unsigned char xdata *) &transmitPacket, HEADER_SIZE_WITHOUT_CRCS);
    transmitPacket.contentCRC = xdataCRC8(contentPtr, contentSize);
    if (sendQueuedPacket()) {
      return PACKET_QUEUED;
    }
    return PACKET_QUEUED;
  }
  return PACKET_NOT_QUEUED;
}

/*
 * Starts transmitting a packet if the channel is clear.  Returns true
 * if packet successfully begins transmitting, returns false otherwise.
 * Note that if the packet can't be sent, it remains in the queue to be
 * sent later.  Therefore, this function should be called occasionly 
 * so as to send stale packets and clear the queue.
 */
unsigned char sendQueuedPacket(void) {
  /*
   * Since this function is called repeatedly in the main loop,
   * it is very important that the two conditions for sending a 
   * packet are checked in the order listed and not in the 
   * reverse order.  In particular, serialState==SCANNING must be
   * checked first, otherwise a packet can be queued twice due 
   * to the way in which packetSent() is implemented.
   */
  if ((serialState == SCANNING) && isPacketInQueue) {
    disableSerialReceive();
    clearTransmitInterruptFlag();
    clearReceiveInterruptFlag();
    enableSerialInterrupts();
    serialState = SENDINGPREAMBLE;
    currentByte = 1;
    disableTimer2();
    byteBuffer = IR_FRAME_START_BYTE;
    SBUF = encodeNybbleTable[0x0F & byteBuffer];
    byteBuffer >>= 4;
    bytesRemaining = 1;
    return TRUE;
  }
  else {
    return FALSE;
  }
}

/*
 * This function is called once a complete packet has been sent.  This
 * function manages all the clean-up necessary after sending a packet 
 * over the communication channel.
 */	
void packetSent(void) {
  isPacketInQueue = FALSE;
  packetSentCallback(&transmitPacket);  // Callback to the next higher level of the communication system.
  startScanning();
  sendQueuedPacket();
}

/*
 * This function is called once a complete packet has been received.  This 
 * function manages the validation of the received packet and passes the 
 * content on to appropriate handlers.
 */
void packetReceived(void) {
  /*
   * Don't want to timeout.
   */
  disableTimer2();
  
  /*
   * If the packet is valid, then deal with it according to its type.
   * Otherwise, enter damage control mode.
   */
  if (receivePacket.headerCRC == xdataCRC8((unsigned char xdata *) &receivePacket, HEADER_SIZE_WITHOUT_CRCS)) {
    if (receivePacket.contentCRC == xdataCRC8(receivePacket.contentPtr, receivePacket.contentSize)) {
      packetReceivedCallback(&receivePacket);
    }
    else {
      contentCRCErrorCallback(&receivePacket);
    }
  }
  else {
    headerCRCErrorCallback();
  }
  
  /*
   * Now that this packet is dealt with, start the communication 
   * subsystem listening for other packets and attempt to send 
   * any packets that may have queued up as a result of receiving
   * this packet.
   */
  startScanning();
  sendQueuedPacket();
}

/*
 * This function configures the processing module for use 
 * with the IR communication module.  Adapted from detectCommModule().
 */
void configureIRCommModule(void) {
  setUARTMode1();		// 8-bit serial communication.
  T2CON = 0x01;			// Configure timer 2 (no auto-reload).
  CKCON |= 0x20;		// Set timer 2 to run off system clock.
  TMOD |= 0x20;			// Set timer1 as 8-bit auto reload.
  PCON |= 0x80;			// Double baud rate; ie, SMOD = 1.
  PRT1CF &= 0x0F;		//|
  IRChannel0 = 1;		//| 
  IRChannel1 = 1;		//| Set pins P1.[4-7] as inputs.
  IRChannel2 = 1;		//|
  IRChannel3 = 1;		//|
  BAUDRATE = BAUD92160;         // Must agree with the resistor and capacitor values on the IR comm module.
}

/*
 * Initiates an interrupt-driven scan and reception of a packet.
 */
void enableCommunication(void) {
  unsigned char i;
  for (i=0; i<IncomingContentBufferSize; i++) {
    IncomingContentBuffer[i] = 0x11;
  }
  for (i=0; i<OutgoingContentBufferSize; i++) {
    OutgoingContentBuffer[i] = 0x22;
  }
  currentByte = 0;
  bytesRemaining = 0;
  enableTimer1();
  clearTimer2InterruptFlag();
  enableTimer2Interrupt();
  clearTransmitInterruptFlag();
  clearReceiveInterruptFlag();
  isPacketInQueue = FALSE;
  enableSerialInterrupts();
  startScanning();
}

/* 
 * Shuts down communication.
 */
void disableCommunication(void) {
  serialState = DISABLED;
  disableTimer1();
  disableTimer2();
  disableTimer2Interrupt();
  disableSerialInterrupts();
  disableSerialReceive();
  clearTimer2InterruptFlag();
  clearTransmitInterruptFlag();
  clearReceiveInterruptFlag();
  isPacketInQueue = FALSE;
}

/*
 * Handles received bytes by building up a packet one byte at a time.  This function 
 * is independent of the physical layer of communication used.  In particular, it is 
 * independent of all interrupts and time insensitive.
 */
void setNextReceivePacketByte(unsigned char c) {
  switch (serialState) {
    case SCANNING : {
      if (c == IR_FRAME_START_BYTE) {
        serialState = LOCKED_ON_FRAME;
      }
      break;
    }
    case LOCKED_ON_FRAME : {
      if (c == IR_FRAME_START_BYTE) {
        break;
      }
      else {
      	/*
      	 * Initialize the receive packet, and set toAddress.
      	 */
      	receivePacket.toAddress = c;
      	receivePacket.fromAddress = 0;
      	receivePacket.packetType = 0;
      	receivePacket.sequenceNumber = 0;
      	receivePacket.contentSize = 0;
      	receivePacket.headerCRC = 0;
      	receivePacket.contentCRC = 0;
      	receivePacket.contentPtr = 0;
      	currentByte = 1;
      	serialState = RECEIVINGHEADER;
      }
    	break;
    }
    case RECEIVINGHEADER : {
      *(((unsigned char xdata *) &receivePacket) + currentByte++) = c;
      if (currentByte >= HEADER_SIZE_WITH_CRCS) {	// Finished receiving the header.  Decide what to do with content.
        // meme : the header CRC is checked twice, once here and once in packetReceived.  should limit this to once.
        if (receivePacket.headerCRC == xdataCRC8((unsigned char xdata *) &receivePacket, HEADER_SIZE_WITHOUT_CRCS)) {
          if ((receivePacket.toAddress == NeighborhoodID) || (receivePacket.toAddress == GLOBALADDRESS))  {
            if (receivePacket.contentSize != 0) {
    		      currentByte = 0;
      		    if (setContentPtr(&receivePacket)) {
      	  		  serialState = RECEIVINGCONTENT; // This packet is meant for this node - start receiving the content.
      		    }
      		    else {
      	  		  serialState = IGNORINGCONTENT; // There's no place to put the incoming content, so ignore it.
      		    }
            }
            else {
      		    serialState = PACKETRECEIVED; // Packet consists only of the header.
            }
          }
          else {
            serialState = IGNORINGCONTENT; // This packet isn't addressed to this node.
          }
        }
        else {
          serialState = PACKETRECEIVED; // Let packetReceived deal with this bad packet.
        }
      }
      break;
    }
    case RECEIVINGCONTENT : {
      /*
       * Received bytes should be stored in extended RAM.
       */
      *(receivePacket.contentPtr + currentByte++) = c;
      if (currentByte >= receivePacket.contentSize) {
        serialState = PACKETRECEIVED;
      }
      break;
    }
    case IGNORINGCONTENT : {
      /*
       * Don't want to tie up the receive buffer by writing ignored data to it.
       */
      currentByte++;
      if (currentByte >= receivePacket.contentSize) {
        // meme : should somehow let packetReceived know that the content of this packet was ignored.
        // for now, do this by corrupting the header CRC.
        receivePacket.headerCRC = ~(receivePacket.headerCRC);

        serialState = PACKETRECEIVED;
      }
      break;
    }
    default : {
      /* 
       * Should never reach this state.  Alert the user.
       */
      flashLED(0xff, 0xffff);
      break;
    }
  }
}

/*
 * Returns the next byte to send out over the communication link.
 */
unsigned char getNextTransmitPacketByte(void) {
  unsigned char xdata next;
  switch (serialState) {
  case SENDINGHEADER :
    next = *(((unsigned char *) &transmitPacket) + (currentByte++));
    if (currentByte >= HEADER_SIZE_WITH_CRCS) {
      currentByte = 0;
      if (transmitPacket.contentSize > 0) {
	serialState = SENDINGCONTENT;
      }
      else {
	serialState = PACKETSENT;
      }
    }				
    break;
  case SENDINGCONTENT :
    next = *(transmitPacket.contentPtr + (currentByte++));
    if (currentByte >= transmitPacket.contentSize) {
      serialState = PACKETSENT;			
    }
    break;
  default :
    break;
  }
  return next;
}

/*
 * Determines where to put the data contained in the incoming packet based on 
 * information contained in the header.  Returns a null pointer if the request 
 * cannot be accommodated for any reason (space limitations, invalid header, etc.).
 */
unsigned char setContentPtr(Packet xdata * packet) {
  if (serialState == RECEIVINGHEADER) {
    if (receivePacket.contentSize <= IncomingContentBufferSize) {
      packet->contentPtr = IncomingContentBuffer;
      return 1;
    }
  }
  return 0;
}


/*
 * Note on the UART:
 * Interrupt number 4 is the serial interrupt.  A serial interrupt is called any time
 * RI or TI is set to 1.  This can be done in either software or hardware.  RI is set 
 * to 1 by the UART hardware whenever the UART has completed receiving a byte.  The 
 * byte just received resides in SBUF.  TI is set to 1 by the UART hardware whenever 
 * the UART has completed transmitting a byte.  Loading a byte into SBUF initiates 
 * the transmission of that byte by the UART.  Thus, the statement SBUF=SBUF instructs 
 * the UART to transmit the byte it last received.  Both TI and RI are only cleared to 
 * 0 in software; the hardware does not clear them at all.
 * 
 * This interrupt service routine checks the state of the transmission or 
 * reception processes and acts accordingly; it either transmits the next byte 
 * or it logs the byte just received and waits for the next byte.  The byteHalfSent 
 * variable keeps tabs on which half of the data byte (high nybble or low nybble) 
 * is being dealt with.  All other global state regarding serial communication is
 * held in the serialState variable.
 */
void serialInterrupt (void) interrupt 4 {
  /* 
   * The receive interrupt (RI) indicates that the UART has completed 
   * receiving a byte.  
   */
  if (RI) {
    clearReceiveInterruptFlag();
    renewReceiveTimeout();
    enableTimer2();
    /* 
     * Two network bytes for each data byte.
     */
    if (!bytesRemaining) {
      byteBuffer = SBUF;
      if ((serialState == SCANNING) && (byteBuffer != encodeNybbleTable[0x0F & IR_FRAME_START_BYTE])) {
        return;
      }
      byteBuffer = decodeNybbleTable[SBUF]; 
      bytesRemaining = 1;
    }
    else {
      byteBuffer |= (decodeNybbleTable[SBUF] << 4);
      bytesRemaining = 0;
      setNextReceivePacketByte(byteBuffer);
    }
    if (serialState == PACKETRECEIVED) {
      packetReceived();
    }
  }
  if (TI) {
    clearTransmitInterruptFlag();
    if (bytesRemaining) {
      /*
       * Send second half of the byte being sent.
       */
      SBUF = encodeNybbleTable[0x0F & byteBuffer];
      bytesRemaining = 0;
    }
    else if (serialState == SENDINGPREAMBLE) {
      /*
       * Start countdown to sending next byte.
       */
      setTimer2Value(TIMER2TRANSMITRELOAD);
      clearTimer2InterruptFlag();
      enableTimer2();
    }
    else if (serialState != PACKETSENT) {
      /*
       * Send the next byte in the packet.
       */
      byteBuffer = getNextTransmitPacketByte();
      SBUF = encodeNybbleTable[0x0F & byteBuffer];
      byteBuffer >>= 4; 
      bytesRemaining = 1;
    }
    else if (serialState == PACKETSENT) {
      packetSent();
    }
  }
}

void timer2Interrupt(void) interrupt 5 {
  /* 
   * Nothing interesting on current channel, so switch to the next channel.
   */
  if (serialState == SCANNING) {
    IRChannelSelect0 = ~IRChannelSelect0;
    if (!IRChannelSelect0) {
      IRChannelSelect1 = ~IRChannelSelect1;
    }
    setTimer2Value(TIMER2RECEIVERELOAD);
  }
  
  /*
   * Wait a while before sending next byte of preamble.
   */
  else if (serialState == SENDINGPREAMBLE) {
    if (currentByte++ < BROADCASTADDRESSREPITITIONS) {
      byteBuffer = IR_FRAME_START_BYTE;
      SBUF = encodeNybbleTable[0x0F & byteBuffer];
      byteBuffer >>= 4; 
      bytesRemaining = 1;
    }
    else {
      serialState = SENDINGHEADER;
      currentByte = 0;
      byteBuffer = getNextTransmitPacketByte();
      SBUF = encodeNybbleTable[0x0F & byteBuffer];
      byteBuffer >>= 4; 
      bytesRemaining = 1;
    }
    disableTimer2();
  }
  
  /*
   * The rest of the packet has gone missing.  Reset communication state.
   */
  else if (REN) {
    currentByte = 0;
    bytesRemaining = 0;
    disableTimer2();
    disableSerialReceive();
    clearReceiveInterruptFlag();
    startScanning();
    // In the days of yore:
    // disableCommunication(); 
    // enableCommunication();
  }
  
  /*
   * Clear the interrupt flag in any case.
   */
  clearTimer2InterruptFlag();
}

/* 
 * Adapted from: http://cell-relay.indiana.edu/mhonarc/cell-relay/1999-Jan/msg00074.html
 * 8 bit CRC Generator, MSB shifted first
 * Polynom: x^8 + x^2 + x^1 + 1
 * 
 * Calculates an 8-bit cyclic redundancy check sum for a buffer of characters in xdata.
 */
unsigned char xdataCRC8(unsigned char xdata * bufferPtr, unsigned int bufferSize) {
  unsigned int xdata i;
  unsigned char xdata crc = 0;
  for (i=0; i < bufferSize; i++) {
    crc = CRC8Table[crc ^ (*(bufferPtr + i))];
  }
  return crc;
}
