/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*************************************************************************
 * ExpansionModules.c
 * target: PushpinV3
 *
 * Last modified 07MAY2002 by lifton.
 *
 * Functions pertaining to Bertha's control of various expansion modules.
 *************************************************************************/

#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>

unsigned int xdata LDRCalibrationConstant = 0;
#define LOW_LEVEL_ADC_NOISE 10

void enableLDRModule(void) {
	/*
	 * Reconfigure crossbar.
	 */
	XBR2 &= ~(0x40);		// Disable crossbar. 
	XBR0 |= 0x28;			//| Output all PCA modules to port pins.
	XBR0 &= ~(0x10);		//|
	XBR2 |= 0x40;			// Enable crossbar.
	/*
	 * Enable LDR and wait for it to settle.  This requires calibrating 
	 * twice for some reason, possibly due to the extra delay.
	 */
	LDR = 1;
	calibrateLDR(LOW_LEVEL_ADC_NOISE);
	/*
	 * Configure PCA.
	 */
	PCA0MD = 0;				// Select system clock /12 for counter.
	PCA0CPM0 = 0x42;		// Set all PCA modules to PWM.
	PCA0CPM1 = 0x42;
	PCA0CPM2 = 0x42;
	PCA0CPM3 = 0x42;
	PCA0CPM4 = 0x42;
	clearLEDDisplay();
	CR = 1;					// Enable PCA counter timer.
}


/*
 * Calibrates the light dependent resistor (LDR) attached to channel
 * one of the ADC.  Samples the ADC up to one hundred times until 
 * the difference between two consecutive samples is less than a 
 * predetermined noise value.  The last sample taken in either case 
 * is taken as the calibration value.  If no two consecutive samples
 * are close enough, the status LED flashes to warn the user.
 *
 * This function could be generalized for all the ADC channels.
 */
void calibrateLDR(unsigned int noise) {
	unsigned char xdata i;
	unsigned int xdata lastSample = getADCValue(1);
	for (i=0; i<100; i++) {
		LDRCalibrationConstant = getADCValue(1);
		if (LDRCalibrationConstant > lastSample) {
			if (LDRCalibrationConstant - lastSample < noise) {
				return;
			}
		}
		else {
			if (lastSample - LDRCalibrationConstant < noise) {
				return;
			}
		}
		lastSample = LDRCalibrationConstant;
	}
	flashLED(100,40000);
}

/*
 * Returns a signed int indicating the value of the LDR relative to 
 * the current LDR calibration constant.  The returned value is 
 * positive if the light reading is brighter than the calibration
 * value, negative if darker.
 */
signed int getLDRValue(void) {
	return ((signed int) getADCValue(1)) - ((signed int) LDRCalibrationConstant);
}

void monitorLDR(unsigned int noise) {
	if (LDRCalibrationConstant > LOW_LEVEL_ADC_NOISE) {
		monitorADCWindow(1, LDRCalibrationConstant + noise, 0xFFFF);
	}
}

/*
 * Sets the duty cycle of the pulse width modulator corresponding 
 * to the LED 'color', based on the value of 'intensity'.  An 
 * intensity of 0xFF corresponds to a zero percent duty cycle, 
 * achieved by actually turning off the PWM instead of setting the 
 * duty cycle to 0xFF, which is low, but non-zero.  The maximum 
 * duty cycle corresponds to an 'intensity' value of 0x00.
 */
void setLEDIntensity(unsigned char color, unsigned char intensity) {
	switch (color) {
		case RED1 :
			if (intensity != 0xFF) {
				PCA0CPM3 = 0x42;
				PCA0CPH3 = intensity;
			}
			else {
				PCA0CPM3 &= ~(0x02);
				PCA0CPM3 |= 0x10;
			}
			break;
		case RED2 :
			if (intensity != 0xFF) {
				PCA0CPM2 = 0x42;
				PCA0CPH2 = intensity;
			}
			else {
				PCA0CPM2 &= ~(0x02);
				PCA0CPM2 |= 0x10;				
			}
			break;
		case AMBER1 :
			if (intensity != 0xFF) {
				PCA0CPM1 = 0x42;
				PCA0CPH1 = intensity;
			}
			else {
				PCA0CPM1 &= ~(0x02);
				PCA0CPM1 |= 0x10;
			}
			break;
		case YELLOW1 :
			if (intensity != 0xFF) {
				PCA0CPM0 = 0x42;
				PCA0CPH0 = intensity;
			}
			else {
				PCA0CPM0 &= ~(0x02);
				PCA0CPM0 |= 0x10;
			}
			break;
		case GREEN1 :
			if (intensity != 0xFF) {
				PCA0CPM4 = 0x42;
				PCA0CPH4 = intensity;
			}
			else {
				PCA0CPM4 &= ~(0x02);
				PCA0CPM4 |= 0x10;
			}
			break;
		default :
			break;
	}
}


/*
 * Turns off all LEDs.
 */
void clearLEDDisplay(void) {
	setLEDIntensity(RED1, 0xFF);
	setLEDIntensity(RED2, 0xFF);
	setLEDIntensity(AMBER1, 0xFF);
	setLEDIntensity(YELLOW1, 0xFF);
	setLEDIntensity(GREEN1, 0xFF);
}

/*
 * A blocking function that flashes the LED of type 'color' located 
 * on the LDR expansion module a certain number of times, at a 
 * certain interval, at a certain intensity, all given as arguments.
 */
void flashLDRLED(unsigned char color, 
                 unsigned int times, 
                 unsigned int interval, 
                 unsigned char intensity) {
	while (times--) {
		setLEDIntensity(color, intensity);
		delay(interval);
		setLEDIntensity(color, 0xFF);
		delay(interval);
	}
}

