/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/***************************************************************************
 * TOF.h
 * target: PushpinTOF
 *
 * Last modified 07JUL2004 by erlkonig.
 *
 * Function prototypes and data structures for the Pushpin TOF codebase.
 *
 ***************************************************************************/

/*
 * These constants determine the (linear) error as
 * a function of measured distance.  This error is likely
 * due to ring up time of the transducer.
 */
#define RING_UP_ERROR_SLOPE ((float) 0.037)      // unitless slope
#define RING_UP_ERROR_INTERCEPT 154              // measured in clock ticks

#define MAX_ELECTMESSAGE_RETRIES 3
#define MAX_ANCHORMESSAGE_RETRIES 1
#define MAX_ROUTEDREQUESTMESSAGE_RETRIES 1
#define MAX_ROUTEDREPLYMESSAGE_RETRIES 3
#define MAX_RANDOM_MESSAGE_DELAY 75000
#define NUMBER_OF_PINGS_TO_STORE 10
#define NUMBER_OF_ROUTING_TABLE_ENTRIES 5
#define NULL_DISTANCE 0xFFFF
#define NULL_COORDINATE 10.0          // An impossible coordinate measurement
#define NULL_HOPCOUNT 0xFF

// 1382400 clock ticks per second
#define ELECTION_LENGTH 2000000
#define ANCHORMESSAGE_DELAY 20000000

enum AnchorTypes {
  ORIGIN, YAXIS, PLUSXHALFPLANE, NOT_AN_ANCHOR
};

// Types of LED color behavior.
enum SweepTypes
{
	SOLID, LINEAR, PARABOLIC, SQUARE, STROBE
};

// LED states that are time-independent; solid colors.
// IMPORTANT!! When new states are added to this enum, LEDSTATE_OFF should remain at the end
enum LEDSolidStates
{
  LEDSTATE_WHITE = 0, LEDSTATE_RED, LEDSTATE_LIME, LEDSTATE_BLUE, 
  LEDSTATE_AQUA, LEDSTATE_FUCHSIA, LEDSTATE_YELLOW, LEDSTATE_MAROON, 
  LEDSTATE_GREEN, LEDSTATE_NAVY, LEDSTATE_TEAL, LEDSTATE_PURPLE, 
  LEDSTATE_OLIVE, LEDSTATE_GRAY, LEDSTATE_SILVER, LEDSTATE_DARKORANGE, LEDSTATE_OFF
};

// LED states that are time-dependent; pulsing and strobing patterns.
enum LEDActiveStates
{ 
  LEDSTATE_REDLINEAR = ( LEDSTATE_OFF + 1), LEDSTATE_REDPARA, LEDSTATE_GREENLINEAR, LEDSTATE_GREENPARA,
  LEDSTATE_BLUELINEAR, LEDSTATE_BLUEPARA, LEDSTATE_PULSE4, LEDSTATE_TOFCHILLIN, LEDSTATE_TOFEVERYONE, 
  LEDSTATE_TOFANCHOR1, LEDSTATE_TOFANCHOR2, LEDSTATE_TOFANCHOR3, LEDSTATE_TOFBOUND1, LEDSTATE_TOFBOUND2,
  LEDSTATE_TEST
};

/* ADC Channels */
enum InputChannels {
  PHOTOTRANSISTOR = 1,  MICROPHONE = 2, MICROPHONE_LOWPASS = 3, SONAR_LOWPASS = 4, TEMP = 8
};

/*
 * Message type definitions
 */
#define TOF_ELECT_MESSAGE 0x55
#define TOF_ANCHOR_MESSAGE 0x56
#define DEBUGGER_REQUEST_MESSAGE 0x77
#define DEBUGGER_DUMP_MESSAGE 0x99
#define ROUTED_REQUEST_MESSAGE 0x78
#define ROUTED_REPLY_MESSAGE 0x79
#define BEGIN_DATA_HARVEST_MESSAGE 0x7A
#define SYN_TEST_MESSAGE 0xAA
#define SET_COLOR_MESSAGE 0x3A
#define SET_LED_STATE_MESSAGE 0x3B


/* Set the MAX and MIN distance for the TOF system
 * The formula to use is:
 *
 * DISTANCE = length_in_meters / (346.65 * 5.425e-07)
 */
#define MAX_TOF_DISTANCE 0x3200
#define MIN_TOF_DISTANCE 0x300


/*
 * Macros
 */

// External Interrupt macro
#define enableLightInterrupt() IE |= 0x01
#define disableLightInterrupt() IE &= ~(0x01)
#define clearLightInterruptFlag() TCON &= ~(0x02)

#define enableSonarInterrupt() IE |= 0x04
#define disableSonarInterrupt() IE &= ~(0x04)
#define clearSonarInterruptFlag() TCON &= ~(0x08)

// Programmable Counter Array macros.
#define enablePCAInterrupts() (EIE1 |= 0x08)
#define disablePCAInterrupts() (EIE1 &= ~(0x08))
#define clearPCAInterruptFlag() (CF = 0)
#define clearPCA3InterruptFlag() (CCF3 = 0)
#define enablePCA3Interrupts() PCA0CPM3 |= 0x01; PCA0CPM3 |= 0x08
#define disablePCA3Interrupts() PCA0CPM3 &= ~(0x01); PCA0CPM3 &= ~(0x08)
#define startPCARunControl() CR = 1
#define stopPCARunControl() CR = 0

/* 
 * Localization information structures
 */

typedef struct PingInfoDefinition { 
  unsigned int localDistance;
  unsigned int minimumKnownDistance;
  unsigned char minimumKnownDistanceID;
  unsigned char anchorID;
  unsigned int anchorDistanceToPinger;
  float anchorXCoordinate;
  float anchorYCoordinate;
  unsigned char minimumHopcountToAnchor;
} PingInfo;

typedef struct RoutingTableEntryDefinition {
  unsigned char nodeID; // The ID of the node from which a request originated.
  unsigned char hopCount;         // The number of hops back to this node.
} RoutingTableEntry;

typedef struct PositionDefinition {
  float x;
  float y;
} Position;


/*
 * Global Variables
 */
extern PingInfo xdata pingInformation[NUMBER_OF_PINGS_TO_STORE];
extern RoutingTableEntry xdata routingTable[NUMBER_OF_ROUTING_TABLE_ENTRIES];
extern Position coordinates;
extern unsigned char xdata elected;
extern unsigned char xdata ledState;   // TOFOutput.c

/* 
 * Time of flight message definition
 */

typedef struct TOFElectMessageDefinition {
  unsigned char type;                  // Message type identifying this as a TOF message.
  unsigned char pingNumber;            // The flash number that corresponds to this message
  unsigned char id;                    // NeighborhoodID of the originating node.
  unsigned int distanceMeasurement;    // TOF Distance measurement.
} TOFElectMessage;

typedef struct TOFAnchorMessageDefinition {
  unsigned char type;                  // Message ID
  unsigned char anchorType;            // Anchor number [0,1,2] 
  unsigned char anchorID;              // Neighborhood ID of the anchor
  unsigned int  distanceToPinger;      // Distance to pinger measured by this anchor
  float xCoordinate;                   // Coordinates of the anchor
  float yCoordinate;
  unsigned char hopCount;              // Hopcount of this message
} TOFAnchorMessage;

typedef struct RoutedRequestMessageDefinition {
  unsigned char type;
  unsigned char requestingNodeID; // The ID of the node to which a reply should be sent
  unsigned char respondingNodeID; // The recipient of this request
  unsigned char hopCount;         // The number of hops back to the requesting node.
} RoutedRequestMessage;

typedef struct RoutedReplyMessageDefinition {
  unsigned char type;
  unsigned char requestingNodeID;
  unsigned char respondingNodeID;
  unsigned char hopCount;
  signed int xCoordinateTicks;
  signed int yCoordinateTicks;
  unsigned int ping1;
  unsigned int ping2;
  unsigned int ping3;
  unsigned int ping4;
  unsigned int ping5;
  unsigned int ping6;
  unsigned int ping7;
  unsigned int ping8;
  unsigned int ping9;
  unsigned int ping10; 
} RoutedReplyMessage;

typedef struct SetColorMessageDefinition {
  unsigned char type;
  unsigned char R;
  unsigned char G;
  unsigned char B;
} SetColorMessage;

typedef struct SetLedStateMessageDefinition {
  unsigned char type;
  unsigned char state;
} SetLedStateMessage;

typedef struct DebuggerRequestMessageDefinition {
  unsigned char type;
} DebuggerRequestMessage;

typedef struct DebuggerDumpMessageDefinition {
  unsigned char type;
  unsigned char id;

  unsigned int ping1LocalDistance;
  unsigned int ping1MinimumKnownDistance;
  unsigned char ping1MinimumKnownDistanceID;
  unsigned char ping1AnchorID;
  unsigned int ping1AnchorDistanceToPinger;
  float ping1AnchorXCoordinate;
  float ping1AnchorYCoordinate;
  unsigned char ping1MinimumHopcountToAnchor;

  unsigned int ping2LocalDistance;
  unsigned int ping2MinimumKnownDistance;
  unsigned char ping2MinimumKnownDistanceID;
  unsigned char ping2AnchorID;
  unsigned int ping2AnchorDistanceToPinger;
  float ping2AnchorXCoordinate;
  float ping2AnchorYCoordinate;
  unsigned char ping2MinimumHopcountToAnchor;

  unsigned int ping3LocalDistance;
  unsigned int ping3MinimumKnownDistance;
  unsigned char ping3MinimumKnownDistanceID;
  unsigned char ping3AnchorID;
  unsigned int ping3AnchorDistanceToPinger;
  float ping3AnchorXCoordinate;
  float ping3AnchorYCoordinate;
  unsigned char ping3MinimumHopcountToAnchor;

  float xCoordinate;
  float yCoordinate;

  unsigned char numOfNeighbors;
} DebuggerDumpMessage;


/*
 * Function Prototypes
 */

/* TOFHardware.c */
void initializeTOFHardware(void);

/* TOFMath.c */
float clockTicksToMeters(unsigned int);
signed int metersToClockTicks(float);
float baselineDistance(float, float);
void computeAnchorCoordinates(unsigned char);
void localizeThisNode(void);

/* TOFMessaging.c */
void initializeTOFMessaging(void);
unsigned char sendTOFElectMessage(unsigned char, unsigned int, unsigned char);
unsigned char sendTOFAnchorMessage(unsigned char, 
                                   unsigned char, 
                                   unsigned int,
                                   float,
                                   float,
                                   unsigned char);

/* TOFInput.c */
void initializeTOFInput(void);
unsigned char getSensorReading(unsigned char);

/* TOFOutput.c */
void initializeTOFOutput( void );
unsigned int TableSqrt( unsigned int );
unsigned char LFO_AmpToHex( unsigned char hexMax, unsigned char scaleType );
void updateLFO( void );
void setLFOFreq( unsigned int LFO_Freq );
void setRGBLEDDisplay( unsigned char RedValue, unsigned char GreenValue, unsigned char BlueValue );
void setRGBLEDPreset( unsigned char preset[] );
void setColorVolume( unsigned char volumeValue );
void setLEDState( unsigned char stateLED, unsigned char volumeLED );
void updateLEDState( void );
void setInputLEDState(unsigned char color[]);

// Meme : Add function prototypes for sending Routed messages
void locateNextNeighbor(void);
void sendQueuedTOFMessage(void);
void removeNodeFromRoutingTable(unsigned char id);
void beginDataHarvest(void);
