/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*******************************************************************************
 * TOFModule.c
 * target: PushpinTOF
 *
 * Last modified 3MAR2004 by mbroxton.
 *
 * Functions for initializing and controllind the hardware on the
 * Pushpin Time of Flight Module.
 *******************************************************************************/
#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>
#include <TOF.h>

void initializeTOFHardware(void) {
  /*
   * Reconfigure crossbar.
   */
  XBR2 &= ~(0x40);		  // Disable crossbar
  XBR0 = 0x04;			  // Enable UART to P0.0 and P0.1
  XBR0 |= 0x18;			  // Output three PCA modules to port pins P0.2-P0.4.
  XBR1 = 0x14;                    // Route /INT0 and /INT1 to port P0.5 and P0.6
  XBR2 |= 0x40;			  // Enable crossbar.
  /* 
   * Configure I/O pins
   */
  PRT0CF |= 0x1C;         // LED pins are push-pull
  PRT0CF &= ~(0x60);      // External interrupt pins are open drain      
  /*
   * Configure DAC to provide voltage reference of 2.4V
   */
  DAC0CN = 0x80;          // Enable the DAC
  DAC0H = 0x0F;		  // Set to max output voltage
  DAC0L = 0xFF;
  // Delay so that the voltage reference value can stabilize
  delay(0xffff);

  /*
   * Configure PCA.
   */
  // Select system clock /12 for counter.
  PCA0MD = 0;				
  // Set three PCA modules to PWM to control intensity of RGB LED.
  PCA0CPM0 = 0x42;
  PCA0CPM1 = 0x42;
  PCA0CPM2 = 0x42;
  clearRGBLEDDisplay();
  /*
   * Set up PCA module 3 as a software timer with interrupt 
   * for use in timing sonar ping.  
   */
  PCA0CPH3 = 0x00;
  PCA0CPL3 = 0x00;
  PCA0CPM3 = 0x48;        // Software Timer mode
  enablePCAInterrupts();
  disablePCA3Interrupts();

  CR = 1;	          // Enable PCA counter timer.

  /*
   * Configure External Interrupts
   */
  disableGlobalInterrupts();  
  TCON |= 0x05;           // Set edge sensitive interrupt for external interrupt lines
  clearLightInterruptFlag();
  enableLightInterrupt();
  enableGlobalInterrupts();
}

/*
 * Sets the duty cycle of the pulse width modulator corresponding 
 * to the LED 'color', based on the value of 'intensity'.  An 
 * intensity of 0xFF corresponds to a zero percent duty cycle, 
 * achieved by actually turning off the PWM instead of setting the 
 * duty cycle to 0xFF, which is low, but non-zero.  The maximum 
 * duty cycle corresponds to an 'intensity' value of 0x00.
 */
void setRGBLEDIntensity(unsigned char color, unsigned char intensity) {
  switch (color) {
  case RGB_RED :
    if (intensity != 0xFF) {
      PCA0CPM0 = 0x42;
      PCA0CPH0 = intensity;
    }
    else {
      PCA0CPM0 &= ~(0x02);
      PCA0CPM0 |= 0x10;
    }
    break;
  case RGB_BLUE :
    if (intensity != 0xFF) {
      PCA0CPM1 = 0x42;
      PCA0CPH1 = intensity;
    }
    else {
      PCA0CPM1 &= ~(0x02);
      PCA0CPM1 |= 0x10;				
    }
    break;
  case RGB_GREEN :
    if (intensity != 0xFF) {
      PCA0CPM2 = 0x42;
      PCA0CPH2 = intensity;
    }
    else {
      PCA0CPM2 &= ~(0x02);
      PCA0CPM2 |= 0x10;
    }
    break;
  default :
    break;
  }
}

// Deactivate all LEDs
void clearRGBLEDDisplay(void) {
  setRGBLEDIntensity(RGB_RED, 0xFF);
  setRGBLEDIntensity(RGB_GREEN, 0xFF);
  setRGBLEDIntensity(RGB_BLUE, 0xFF);
}
