/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/***************************************************************************
 * TOFMath.c
 * target: PushpinTOF
 *
 * Last modified 14MAR2004 by mbroxton.
 *
 * Funtions for computing location based on a number of time of flight 
 * measurements from the camera flash/sonar pinger.  
 ***************************************************************************/

#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>
#include <TOF.h>
// SDCC Math Library
#include <math.h>

/*
 * For any pair of distance measurements between the
 * pinger and two nodes, this function computes the 
 * baseline distance between the two nodes assuming 
 * that the pinger was held directly above one of them 
 * ( thus forming a right triangle ).
 *
 * Precise Explanation:
 * Given a right triangle with two sides 'a' and 'b'
 * and hypotenuse 'c', this function computes the 
 * side b given sides a and c.  
 */
float baselineDistance(float leg, float hypotenuse) {
  if (hypotenuse > leg) {
    return sqrtf(hypotenuse*hypotenuse - leg*leg);
  } else {
    return 0;
  }
}


void computeAnchorCoordinates(unsigned char anchorType) {
  float xdata d1, d2, d3;
 // float xdata b12, b23, b31;  // bnm = distance between anchor n and anchor m

  switch (anchorType) {
    case ORIGIN : {
      coordinates.x = 0;
      coordinates.y = 0;
      break;
    }
    case YAXIS : {
      coordinates.x = 0;
      coordinates.y = baselineDistance(clockTicksToMeters(pingInformation[ORIGIN].anchorDistanceToPinger),
                                       clockTicksToMeters(pingInformation[ORIGIN].localDistance));
      break;
    }
    case PLUSXHALFPLANE : {
      d1 = pingInformation[YAXIS].anchorYCoordinate;
      d2 = baselineDistance( clockTicksToMeters(pingInformation[YAXIS].anchorDistanceToPinger),
                             clockTicksToMeters(pingInformation[YAXIS].localDistance));
      d3 = baselineDistance( clockTicksToMeters(pingInformation[ORIGIN].anchorDistanceToPinger),
                             clockTicksToMeters(pingInformation[ORIGIN].localDistance));
      coordinates.y = ((d1/2) - ((d2*d2) / (2*d1)) + ((d3*d3)/(2*d1)));
      coordinates.x = sqrtf(fabsf( (d3*d3)-((coordinates.y)*(coordinates.y)) ));
      
      break;
    }
    default : {
      break;
    }
  }
}

/*
 * Meme: Add a good comment for this function once I can type 
 * faster.
 *
 * This function must be declared reentrant so that its parameters 
 * and local variables are located on the stack.  If it is not,
 * the compiler will allocate too many sloc variables in data 
 * memory, which will cause the data memory to overflow.
 */
void localizeThisNode(void) reentrant {


  static float denominator;
  static float sum1;
  static float sum2;
  static float da; 
  static float db; 
  static float dc;

  #define xa (pingInformation[ORIGIN].anchorXCoordinate)
  #define ya (pingInformation[ORIGIN].anchorYCoordinate)
  #define xb (pingInformation[YAXIS].anchorXCoordinate)
  #define yb (pingInformation[YAXIS].anchorYCoordinate)
  #define xc (pingInformation[PLUSXHALFPLANE].anchorXCoordinate)
  #define yc (pingInformation[PLUSXHALFPLANE].anchorYCoordinate)

  da = baselineDistance( clockTicksToMeters(pingInformation[ORIGIN].anchorDistanceToPinger),
                         clockTicksToMeters(pingInformation[ORIGIN].localDistance));
  db = baselineDistance( clockTicksToMeters(pingInformation[YAXIS].anchorDistanceToPinger),
                         clockTicksToMeters(pingInformation[YAXIS].localDistance));
  dc = baselineDistance( clockTicksToMeters(pingInformation[PLUSXHALFPLANE].anchorDistanceToPinger),
                         clockTicksToMeters(pingInformation[PLUSXHALFPLANE].localDistance));

  denominator = 4 * (( xa - xc ) * ( yb - yc ) - ( ya - yc ) * ( xb - xc ));

  sum1 = ( xa * xa ) - ( xc * xc ) + ( ya * ya ) - ( yc * yc ) + ( dc * dc ) - ( da * da );

  sum2 = ( xb * xb ) - ( xc * xc ) + ( yb * yb ) - ( yc * yc ) + ( dc * dc ) - ( db * db );

  coordinates.x = 2 * ( yb - yc ) * sum1 + -2 * ( ya - yc ) * sum2;
  coordinates.y = -2 * ( xb - xc ) * sum1 + 2 * ( xa - xc ) * sum2;

  coordinates.x /= denominator;
  coordinates.y /= denominator;
}

/*
 * There are 5317 clock ticks per meter 
 */
float clockTicksToMeters(unsigned int n) {
  return (float)n / (float)5317.0;
}

signed int metersToClockTicks(float n) {
  return (signed int) (n * 5317.0);
}
