/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/****************************************************************************
 * TOFMessaging.c
 * target: PushpinTOF
 *
 * Last modified 06JUL2004 by erlkonig.
 *
 * Functions, structures, and variables associated with the TOF Module's 
 * communication and messagind protocols.
 ***************************************************************************/
#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>
#include <TOF.h>

TOFElectMessage xdata tofElectMessage;
TOFAnchorMessage xdata tofAnchorMessage;
DebuggerDumpMessage xdata ideMessage;
RoutedRequestMessage xdata routedRequestMessage;
RoutedReplyMessage xdata routedReplyMessage;
unsigned char xdata messageRetries;      // Number of remaining tranmit attempts for this message.
unsigned char xdata messageQueued;       // 1 if a message is queued in the messaging system.
unsigned char xdata queuedPacketType;    // The type of the currently queued message.
unsigned char xdata delayedElectMessageNeedsToBeSent; // Semaphore to prevent the loss
unsigned char xdata delayedElectMessagePingNumber;    //  of an elect message.

unsigned char xdata currentNeighborIndex; 
unsigned char xdata currentRoutingTableIndex;
unsigned long xdata lastMessageTransmitTime;  
unsigned long xdata randomMessageDelay;

/*
 * The routing table maintains the information necessary to 
 * send messages "up the gradient" to the originator of a message
 * request.
 */
RoutingTableEntry xdata routingTable[NUMBER_OF_ROUTING_TABLE_ENTRIES];


void initializeTOFMessaging(void) {
  unsigned char i;
  messageQueued = 0;
  messageRetries = 0;
  lastMessageTransmitTime = 0; 
  delayedElectMessageNeedsToBeSent = FALSE;
  delayedElectMessagePingNumber = 0;
  currentNeighborIndex = 0;
  randomMessageDelay = 0x8FF;

  currentRoutingTableIndex = 0;
  for (i = 0; i < NUMBER_OF_ROUTING_TABLE_ENTRIES ; i++) {
    routingTable[i].nodeID = GLOBALADDRESS;
    routingTable[i].hopCount = NULL_HOPCOUNT;
  }
}


/* 
 * Change the value of the global variable, currentNeighborID
 * to the ID of the next neighbor in the table.  If there are
 * no more neighbors, stop at MAX_NUM_OF_NEIGHBORS 
 */
void locateNextNeighbor(void) {
  if (currentNeighborIndex >= MAX_NUM_OF_NEIGHBORS) {
    currentNeighborIndex = 0;
    if (getNeighborID(currentNeighborIndex) != GLOBALADDRESS) {
      return;
    }
  }
  while ((getNeighborID(++currentNeighborIndex) == GLOBALADDRESS) &&
         (currentNeighborIndex < MAX_NUM_OF_NEIGHBORS));
}         


/*
 * At a random time interval dictated by MAX_RANDOM_MESSAGE_DELAY,
 * attempt to send any message which has been queued in the TOF 
 * message system.  
 */
void sendQueuedTOFMessage(void) {
  /* 
   * Attempt to send a packet only occasionally 
   */
  if ( (getTime() - lastMessageTransmitTime) > randomMessageDelay) {

    // Reset the randomizer with a new random value.
    lastMessageTransmitTime = getTime();
    randomMessageDelay = ((unsigned long)((rand() & 0xFF) * MAX_RANDOM_MESSAGE_DELAY)) >> 8;

    /* 
     * If a TOF message has been queued, the attempt to send it as a
     * broadcast packet.  
     */
    if ((messageRetries > 0) && (messageQueued)) {
      switch (queuedPacketType) {
        case TOF_ELECT_MESSAGE : {
          if (queuePacket(GLOBALADDRESS, 
                          BROADCAST, 
                          sizeof(TOFElectMessage), 
                          (unsigned char xdata*)(&tofElectMessage)) == PACKET_QUEUED) {
            messageRetries--;
          }
          break;
        }
        case TOF_ANCHOR_MESSAGE : {
          if (queuePacket(GLOBALADDRESS, 
                          BROADCAST, 
                          sizeof(TOFAnchorMessage), 
                          (unsigned char xdata*)(&tofAnchorMessage)) == PACKET_QUEUED) {
            messageRetries--;
          }
          break;
        }
        case ROUTED_REQUEST_MESSAGE : {
          if (queuePacket(GLOBALADDRESS, 
                          BROADCAST, 
                          sizeof(routedRequestMessage), 
                          (unsigned char xdata*)(&routedRequestMessage)) == PACKET_QUEUED) {
            messageRetries--;
          }
          break;
        }
        case ROUTED_REPLY_MESSAGE : {
          if (queuePacket(GLOBALADDRESS, 
                          BROADCAST, 
                          sizeof(routedReplyMessage), 
                          (unsigned char xdata*)(&routedReplyMessage)) == PACKET_QUEUED) {
            messageRetries--;
          }
          break;
        }
        default : {
          /* Should never get here. */
          flashLED(0xFF, 0xFFFF);
          break;
        }
      }
    } else {
      /*
       * We have finished transmitting the message.  Reset
       * the message queue check to see if there is an Elect
       * message which was delayed and still needs to be sent.
       */
      messageQueued = 0;
      if (delayedElectMessageNeedsToBeSent) {
        sendTOFElectMessage(delayedElectMessagePingNumber,
                            pingInformation[delayedElectMessagePingNumber].minimumKnownDistance,
                            pingInformation[delayedElectMessagePingNumber].minimumKnownDistanceID);
        delayedElectMessageNeedsToBeSent = FALSE;
      }
    }    
  }
}


/* 
 * Queue a TOF message to be transmitted using the Bertha
 * packet communication system.  The actual transmission
 * of the message itself is handled outside of the interrupt
 * routines in TOFRunLoop(), which is part of the main bertha loop.  
 *
 * This function returns 1 if the message was succesfully queued.
 * Returns 0 if another message was already queued.  
 */
unsigned char sendTOFElectMessage(unsigned char pingNumber, 
                                  unsigned int measurement, 
                                  unsigned char id) {
  if (!messageQueued) {
    tofElectMessage.type = TOF_ELECT_MESSAGE;
    tofElectMessage.pingNumber = pingNumber;
    tofElectMessage.distanceMeasurement = measurement;
    tofElectMessage.id = id;

    messageRetries = MAX_ELECTMESSAGE_RETRIES;  
    queuedPacketType = TOF_ELECT_MESSAGE;
    messageQueued = 1;
    return messageQueued;
  }

  /* 
   * Another message was already queued. 
   */
  return 0;
}

/* 
 * Queue a TOF anchor message to be transmitted using the Bertha
 * packet communication system.  The actual transmission
 * of the message itself is handled outside of the interrupt
 * routines in TOFRunLoop(), which is part of the main bertha loop.  
 *
 * This function returns 1 if the message was succesfully queued.
 * Returns 0 if another message was already queued.  
 */
unsigned char sendTOFAnchorMessage(unsigned char aType, 
                                   unsigned char aID, 
                                   unsigned int aDistanceToPinger,
                                   float aXCoordinate,
                                   float aYCoordinate,
                                   unsigned char hopCount) {
  if (!messageQueued) {
    tofAnchorMessage.type = TOF_ANCHOR_MESSAGE;
    tofAnchorMessage.anchorType = aType;
    tofAnchorMessage.anchorID = aID;
    tofAnchorMessage.distanceToPinger = aDistanceToPinger;
    tofAnchorMessage.xCoordinate = aXCoordinate;
    tofAnchorMessage.yCoordinate = aYCoordinate;
    tofAnchorMessage.hopCount = hopCount;

    messageRetries = MAX_ANCHORMESSAGE_RETRIES;  
    queuedPacketType = TOF_ANCHOR_MESSAGE;
    messageQueued = 1;
    return messageQueued;
  }

  /* 
   *Another message was already queued. 
   */
  return 0;
}

/* 
 * Queue a broadcast request message to be transmitted using the Bertha
 * packet communication system.  The actual transmission
 * of the message itself is handled outside of the interrupt
 * routines in TOFRunLoop(), which is part of the main bertha loop.  
 *
 * This function returns 1 if the message was succesfully queued.
 * Returns 0 if another message was already queued.  
 */
unsigned char sendRoutedRequestMessage(unsigned char requestingNodeID, 
                                       unsigned char respondingNodeID,
                                       unsigned char hopCount) {

  if (!messageQueued) {
    routedRequestMessage.type = ROUTED_REQUEST_MESSAGE;
    routedRequestMessage.requestingNodeID = requestingNodeID;
    routedRequestMessage.respondingNodeID = respondingNodeID;
    routedRequestMessage.hopCount = hopCount;

    messageRetries = MAX_ROUTEDREQUESTMESSAGE_RETRIES;  
    queuedPacketType = ROUTED_REQUEST_MESSAGE;
    messageQueued = 1;
    return messageQueued;
  }

  /* 
   *Another message was already queued. 
   */
  return 0;
}

/* 
 * Queue a broadcast reply message to be transmitted using the Bertha
 * packet communication system.  The actual transmission
 * of the message itself is handled outside of the interrupt
 * routines in TOFRunLoop(), which is part of the main bertha loop.  
 *
 * This function returns 1 if the message was succesfully queued.
 * Returns 0 if another message was already queued.  
 */
unsigned char sendRoutedReplyMessage(unsigned char requestingNodeID, 
                                     unsigned char respondingNodeID,
                                     unsigned char hopCount,
                                     signed int xCoordinate,
                                     signed int yCoordinate) {
  if (!messageQueued) {
    routedReplyMessage.type = ROUTED_REPLY_MESSAGE;
    routedReplyMessage.requestingNodeID = requestingNodeID;
    routedReplyMessage.respondingNodeID = respondingNodeID;
    routedReplyMessage.hopCount = hopCount;
    routedReplyMessage.xCoordinateTicks = xCoordinate;
    routedReplyMessage.yCoordinateTicks = yCoordinate;

    routedReplyMessage.ping1 = pingInformation[0].localDistance;
    routedReplyMessage.ping2 = pingInformation[1].localDistance;
    routedReplyMessage.ping3 = pingInformation[2].localDistance;
    routedReplyMessage.ping4 = pingInformation[3].localDistance;
    routedReplyMessage.ping5 = pingInformation[4].localDistance;
    routedReplyMessage.ping6 = pingInformation[5].localDistance;
    routedReplyMessage.ping7 = pingInformation[6].localDistance;
    routedReplyMessage.ping8 = pingInformation[7].localDistance;
    routedReplyMessage.ping9 = pingInformation[8].localDistance;
    routedReplyMessage.ping10 = pingInformation[9].localDistance;

    messageRetries = MAX_ROUTEDREPLYMESSAGE_RETRIES;  
    queuedPacketType = ROUTED_REPLY_MESSAGE;
    messageQueued = 1;
    return messageQueued;
  }

  /* 
   *Another message was already queued. 
   */
  return 0;
}


/*
 * RECEIVE MESSAGE ROUTINES
 *
 */


void receiveTOFElectMessage(TOFElectMessage* electMessage) {
  /* 
   * Is this new distance measurement smaller than 
   * those we have seen before it for this ping?
   */
  if ( electMessage->distanceMeasurement < pingInformation[electMessage->pingNumber].minimumKnownDistance ) {
    pingInformation[electMessage->pingNumber].minimumKnownDistance = electMessage->distanceMeasurement;
    pingInformation[electMessage->pingNumber].minimumKnownDistanceID = electMessage->id;

    // Color ourselves "not it".
    setLEDState( LEDSTATE_TOFEVERYONE, 0xFF );
        
    // If we received this message, we must not be elected. 
    elected = 0;

    // Pass the message along to other nodes.
    if ( !(sendTOFElectMessage(electMessage->pingNumber,
                               electMessage->distanceMeasurement,
                               electMessage->id
                               ))) {
      /* 
       * If another message is being sent, set a semaphore
       * to indicate that this message needs to be sent once 
       * the current message has completed trasmission.
       */
      delayedElectMessageNeedsToBeSent = TRUE;
      delayedElectMessagePingNumber = electMessage->pingNumber;
    }
  }
}

void receiveTOFAnchorMessage(TOFAnchorMessage* anchorMessage) {
  /* 
   * Record information from the anchor packet.
   */
  pingInformation[anchorMessage->anchorType].anchorID = anchorMessage->anchorID;
  pingInformation[anchorMessage->anchorType].anchorDistanceToPinger = anchorMessage->distanceToPinger;
  pingInformation[anchorMessage->anchorType].anchorXCoordinate = anchorMessage->xCoordinate;
  pingInformation[anchorMessage->anchorType].anchorYCoordinate = anchorMessage->yCoordinate;     
  
  /*
   * If the hopcount of this message is <= to the previously recorded hop 
   * count, retransmit this message.
   */
  if ( anchorMessage->hopCount <=
       pingInformation[anchorMessage->anchorType].minimumHopcountToAnchor) {
    pingInformation[anchorMessage->anchorType].minimumHopcountToAnchor = anchorMessage->hopCount;
    sendTOFAnchorMessage(anchorMessage->anchorType, 
                         anchorMessage->anchorID,
                         anchorMessage->distanceToPinger,
                         anchorMessage->xCoordinate,
                         anchorMessage->yCoordinate,
                         anchorMessage->hopCount + 1);
  } 
}


void receiveDebuggerRequestMessage() {
      ideMessage.type = DEBUGGER_DUMP_MESSAGE;
      ideMessage.id = NeighborhoodID;
      
      ideMessage.ping1LocalDistance = pingInformation[0].localDistance;
      ideMessage.ping1MinimumKnownDistance = pingInformation[0].minimumKnownDistance;
      ideMessage.ping1MinimumKnownDistanceID = pingInformation[0].minimumKnownDistanceID;
      ideMessage.ping1AnchorID = pingInformation[0].anchorID;
      ideMessage.ping1AnchorDistanceToPinger = pingInformation[0].anchorDistanceToPinger;
      ideMessage.ping1AnchorXCoordinate = pingInformation[0].anchorXCoordinate;
      ideMessage.ping1AnchorYCoordinate = pingInformation[0].anchorYCoordinate;
      ideMessage.ping1MinimumHopcountToAnchor = pingInformation[0].minimumHopcountToAnchor;

      ideMessage.ping2LocalDistance = pingInformation[1].localDistance;
      ideMessage.ping2MinimumKnownDistance = pingInformation[1].minimumKnownDistance;
      ideMessage.ping2MinimumKnownDistanceID = pingInformation[1].minimumKnownDistanceID;
      ideMessage.ping2AnchorID = pingInformation[1].anchorID;
      ideMessage.ping2AnchorDistanceToPinger = pingInformation[1].anchorDistanceToPinger;
      ideMessage.ping2AnchorXCoordinate = pingInformation[1].anchorXCoordinate;
      ideMessage.ping2AnchorYCoordinate = pingInformation[1].anchorYCoordinate;
      ideMessage.ping2MinimumHopcountToAnchor = pingInformation[1].minimumHopcountToAnchor;

      ideMessage.ping3LocalDistance = pingInformation[2].localDistance;
      ideMessage.ping3MinimumKnownDistance = pingInformation[2].minimumKnownDistance;
      ideMessage.ping3MinimumKnownDistanceID = pingInformation[2].minimumKnownDistanceID;
      ideMessage.ping3AnchorID = pingInformation[2].anchorID;
      ideMessage.ping3AnchorDistanceToPinger = pingInformation[2].anchorDistanceToPinger;
      ideMessage.ping3AnchorXCoordinate = pingInformation[2].anchorXCoordinate;
      ideMessage.ping3AnchorYCoordinate = pingInformation[2].anchorYCoordinate;
      ideMessage.ping3MinimumHopcountToAnchor = pingInformation[2].minimumHopcountToAnchor;

      ideMessage.xCoordinate = coordinates.x;
      ideMessage.yCoordinate = coordinates.y;

      ideMessage.numOfNeighbors = getNumOfNeighbors();
      
      queuePacket(IDEADDRESS, 
                  BROADCAST, 
                  sizeof(DebuggerDumpMessage), 
                  (unsigned char xdata *) &ideMessage);
}

void removeNodeFromRoutingTable(unsigned char id) {
  unsigned char i;
  for (i=0; i<NUMBER_OF_ROUTING_TABLE_ENTRIES; i++) {
    if (routingTable[i].nodeID == id) {
      routingTable[i].nodeID = GLOBALADDRESS;
      routingTable[i].hopCount = NULL_HOPCOUNT;
      break;
    }
  }
}

void updateRoutingTable(RoutedRequestMessage * message) {
  unsigned char i = 0;
  unsigned char nodeID;
  /*
   * Determine which node is the origin of this gradient.
   */
  if (message->type == ROUTED_REQUEST_MESSAGE) {
    nodeID = message->requestingNodeID;
  }
  else if (message->type == ROUTED_REPLY_MESSAGE) {
    nodeID = message->respondingNodeID;
  }
  else {
    return;
  }
  /*
   * Attempt to locate the entry in the table.
   */
  while (i < NUMBER_OF_ROUTING_TABLE_ENTRIES) {
    if (routingTable[i].nodeID == nodeID) {
      /*
       * Adjust existing entry only if the new message has a lower
       * hopCount.  A lower hopCount signifies a more efficient path
       * to the requestingNodeID.
       */
      if (message->hopCount < routingTable[i].hopCount) {
        routingTable[i].hopCount = message->hopCount;
      }
      return;
    } 
    i++;
  }
  
  /*
   * If the routing table does not contain an entry for this requestingNodeID,
   * add a new entry to the table
   */
  routingTable[currentRoutingTableIndex].nodeID = nodeID;
  routingTable[currentRoutingTableIndex].hopCount = message->hopCount;

  /*
   * Increment the index of the routing table counter.
   * Wrap around to the first entry if the table is full. 
   * (The routing table is FIFO)
   */
  if (++currentRoutingTableIndex >= NUMBER_OF_ROUTING_TABLE_ENTRIES) {
    currentRoutingTableIndex = 0;
  }
}

/*
 * Search for a route in the routing table.  If a 
 * route is found to requestingNodeID, we return the hopCount
 * of our route.  Otherwise, we return NULL_HOPCOUNT.
 */
unsigned char getHopCountForRoute(unsigned char nodeID) {
  unsigned char i = 0;

  /*
   * Attempt to locate the entry in the table.
   * If we find it, return the hopcount.
   */
  while (i < NUMBER_OF_ROUTING_TABLE_ENTRIES) {
    if (routingTable[i].nodeID == nodeID) {
      return routingTable[i].hopCount;
    } 
    i++;
  }
  /* If the route is not in the table, return NULL_HOPCOUNT */
  return NULL_HOPCOUNT;
}


void  receiveRoutedRequestMessage(RoutedRequestMessage* message) {
  /*
   * Update the routing table with the information in this packet.
   */
  updateRoutingTable(message);
  
  if (NeighborhoodID != message->respondingNodeID) {
    setLEDState( LEDSTATE_RED, 0xFF );
  }

  if (message->hopCount <= getHopCountForRoute(message->requestingNodeID)) {
    if (message->respondingNodeID == NeighborhoodID) {

      /* meme - PROCESS MESSAGE */
      setLEDState( LEDSTATE_GREEN, 0xFF );

      flashLED(2, 0xffff); // delay between sending message so dongle can keep up.
      sendRoutedReplyMessage(message->requestingNodeID,
                             NeighborhoodID,
                             0,
                             metersToClockTicks(coordinates.x),
                             metersToClockTicks(coordinates.y));
    } else {
      /*
       * Rebroadcast the message to other nodes.
       */
      sendRoutedRequestMessage(message->requestingNodeID, message->respondingNodeID, message->hopCount + 1);
    }
  }
}

void receiveRoutedReplyMessage(RoutedReplyMessage *message) {

  if (NeighborhoodID != message->respondingNodeID) {
    setLEDState( LEDSTATE_BLUE, 0xFF );
  }
  /*
   * Update the routing table with the information in this packet.
   */
  updateRoutingTable((RoutedRequestMessage *) message);
  removeNodeFromRoutingTable(IDEADDRESS);

  if (message->hopCount <= getHopCountForRoute(message->respondingNodeID)) {
    if (message->requestingNodeID == NeighborhoodID) {
      /*
       * Don't do anything.
       */
    } else {
      /*
       * Rebroadcast the message to other nodes.
       */
      sendRoutedReplyMessage(message->requestingNodeID, 
                             message->respondingNodeID, 
                             message->hopCount + 1,
                             message->xCoordinateTicks,
                             message->yCoordinateTicks);
    }
  }
}

/* 
 * Receives a message that contains RGB values and sets the LED to those values
 */
void receiveSetColorMessage(SetColorMessage *message) {
  unsigned char color[3] = { message->R, message->G, message->B };
  //  setInputLEDState(color);
}

/* 
 * Meme This is not what i want
 */
void receiveSetLedStateMessage(SetLedStateMessage *message) {
  // setLEDState(message->state,0xFF);
}

/*
 * Process an incoming communication message by delegating it
 * to the appropriate receive routine.
 */
void receiveTOFmessage(Packet* packetPtr) {

  switch (((TOFElectMessage *)(packetPtr->contentPtr))->type) {
    case TOF_ELECT_MESSAGE : {
      receiveTOFElectMessage((TOFElectMessage *)(packetPtr->contentPtr));
      break;
    }
    case TOF_ANCHOR_MESSAGE : {
      receiveTOFAnchorMessage((TOFAnchorMessage *)(packetPtr->contentPtr));
      break;
    }
    case DEBUGGER_REQUEST_MESSAGE : {
      receiveDebuggerRequestMessage();
      break;
    }
    case ROUTED_REQUEST_MESSAGE : {
      receiveRoutedRequestMessage((RoutedRequestMessage *)(packetPtr->contentPtr));
      break;
    }
    case ROUTED_REPLY_MESSAGE : {
      receiveRoutedReplyMessage((RoutedReplyMessage *)(packetPtr->contentPtr));
      break;
    }
    case BEGIN_DATA_HARVEST_MESSAGE : {
      beginDataHarvest();
      break;
    }
    case SET_COLOR_MESSAGE : { 
      receiveSetColorMessage((SetColorMessage *)(packetPtr->contentPtr));
      break;
    }
    case SET_LED_STATE_MESSAGE : { 
      receiveSetLedStateMessage((SetLedStateMessage *)(packetPtr->contentPtr));
      break;
    }
    default : {
      break;
    }
  }
}
