/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/**************************************************************************
 * TOFModule.c
 * target: PushpinTOF
 *
 * Josh Lifton, Michael Broxton, Daniel Alejandro Benitez
 * MIT Media Lab
 * Copyright 2002 all rights reserved.
 *
 * Last modified 06JUL2004 by erlkonig.
 *
 * Functions pertaining to Bertha's control of the time of flight
 * expansion module.
 *************************************************************************/
#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>
#include <TOF.h>
/*
 * Function Prototypes 
 */
void lineDemo(void);

/* 
 * Global Variables 
 */
PingInfo xdata pingInformation[NUMBER_OF_PINGS_TO_STORE];
unsigned char xdata currentPing;         // The number of the ping _which we are expecting_
unsigned char xdata anchorType;          // Determines whether this node is an anchor, and if so, which one.
unsigned long xdata anchorTransmitTimer;
unsigned long xdata electionStartTime;   // Start time of the election.  Should be 0 if no election is running.
unsigned char xdata elected;             // 0 if we have been un-elected in receiveTOFMessage()
Position xdata coordinates;
float lineBoundary;
unsigned long checkLineTime;
unsigned char harvestingData;

void beginDataHarvest(void) {
  harvestingData = 1;
}

void initializeTOFModule(void) {
  unsigned char i;

  /* 
   * Initialize global variables
   */
  lineBoundary = 0.0;
  checkLineTime = 0;
  harvestingData = 0;

  coordinates.x = NULL_COORDINATE;
  coordinates.y = NULL_COORDINATE;

  anchorType = NOT_AN_ANCHOR;
  anchorTransmitTimer = 0;
  electionStartTime = 0;

  currentPing = 0;
  for (i = 0 ; i < NUMBER_OF_PINGS_TO_STORE; i++) {
    pingInformation[i].localDistance = NULL_DISTANCE;
    pingInformation[i].minimumKnownDistanceID = GLOBALADDRESS;
    pingInformation[i].minimumKnownDistance = NULL_DISTANCE;
    pingInformation[i].anchorID = GLOBALADDRESS;
    pingInformation[i].anchorDistanceToPinger = NULL_DISTANCE;
    pingInformation[i].anchorXCoordinate = NULL_DISTANCE;
    pingInformation[i].anchorYCoordinate = NULL_DISTANCE;
    pingInformation[i].minimumHopcountToAnchor = 0xFF;
  }

  /*
   * Initialize the Time of Flight module hardware and messaging
   * system and activate the interrupts.
   */
  initializeTOFMessaging();
  initializeTOFHardware();
  initializeTOFInput();
  initializeTOFOutput();
}


void performAnchorDuties(void) {
  // only maintain color if not all pings have been received
  if (!((pingInformation[0].anchorID != GLOBALADDRESS) &&
    (pingInformation[1].anchorID != GLOBALADDRESS) &&
    (pingInformation[2].anchorID != GLOBALADDRESS))) {
    switch (anchorType) {
      case ORIGIN : {
	setLEDState( LEDSTATE_WHITE, 0xFF );
        break;
      }
      case YAXIS : { 
	setLEDState( LEDSTATE_WHITE, 0xFF );
        break;
      }
      case PLUSXHALFPLANE : {
	setLEDState( LEDSTATE_WHITE, 0xFF );
        break;
      }
      default : {
        /* Should never get here. */
	flashLED(0xFF, 0xFFFF);
	break;
      } 
    }
  }
     
  /* 
   * If this node is an anchor, send out an occasional anchor
   * packet.
   */
  if ( (getTime() - anchorTransmitTimer) > ANCHORMESSAGE_DELAY ) {
        anchorTransmitTimer = getTime();
        sendTOFAnchorMessage(anchorType, 
                             NeighborhoodID,
                             pingInformation[anchorType].localDistance,
                             coordinates.x,
                             coordinates.y,
                             0); 
  }
}

/*
 * This is the first demo showing localized Pushpins in action.
 * We color ourselves according to a gradient which changes from 
 * red to green as the coordinate of the node changes from 
 * +y to -y and red to blue as we move from +x to -x.
 */
void gradientDemo(void) {
  unsigned int red = (unsigned char)(0xFF * ((coordinates.x + 0.1) / 0.2)) + 
                     (unsigned char)(0xFF * ((coordinates.y + 0.1) / 0.2));
  
  // Color ourselves "not it".
  setRGBLEDIntensity(RGB_BLUE, 1 - (unsigned char)(0xFF * ((coordinates.x + 0.1) / 0.2)));
  setRGBLEDIntensity(RGB_GREEN, 1 - (unsigned char)(0xFF * ((coordinates.y + 0.1) / 0.2)));
  setRGBLEDIntensity(RGB_RED, red >> 2);
}

void lineDemo(void) { 
  if (anchorType == NOT_AN_ANCHOR) {
    if (getTime() - checkLineTime > 100000) { // increment once every 1/10 second.
      checkLineTime = getTime();
      lineBoundary += 0.01;
      if (lineBoundary > 1.0) {
        lineBoundary = -1.0;
      }
      else if (lineBoundary < -1.0) {
        lineBoundary = 1.0;
      }
      if (coordinates.x > lineBoundary) {
	setLEDState( LEDSTATE_RED, 0xFF );
      } else {
	setLEDState( LEDSTATE_GREEN, 0xFF );
      }
    }
  }
}

/*
 * This is the TOF function that is called from the main 
 * Bertha runloop.  This function can handle tasks which 
 * should not be run from inside the interrupt routines.
 */ 
void TOFRunLoop(void) {
  /* 
   * Send any pending packets
   */
  sendQueuedTOFMessage();

  /* 
   * Update the LED based on the current ledState. 
   */  
   updateLEDState();

  /*
   * Check to see if the election timer has run out.  If it has,
   * and we have not encountered a ElectMessage with a lower 
   * distance measurement, then we assume that we are the origin,
   * the y-axis, or the node in the +x halfplane.
   */
  if ( (electionStartTime > 0) && (getTime() - electionStartTime > ELECTION_LENGTH) ) {
    electionStartTime = 0;

   
    /* 
     * If we are elected, set our anchor ID and compute our coordinates.
     */
    if (elected) {
      anchorType = currentPing - 1;  
      computeAnchorCoordinates(anchorType);
    }    
  }

  /*
   * If we are an elected anchor point, we must tend to our duties
   */
  if (anchorType != NOT_AN_ANCHOR && !harvestingData) {
    performAnchorDuties();
  }
  

  /*
   * Check to see if a solution has been found. If a solution
   * is available, we heard all three pings, and we are not an 
   * anchor, compute the solution!
   */
   
  if ((pingInformation[0].anchorID != GLOBALADDRESS) &&
      (pingInformation[1].anchorID != GLOBALADDRESS) &&
      (pingInformation[2].anchorID != GLOBALADDRESS)) {
    if ((pingInformation[0].localDistance != NULL_DISTANCE) &&
      (pingInformation[1].localDistance != NULL_DISTANCE) &&
      (pingInformation[2].localDistance != NULL_DISTANCE)) {
      if ((anchorType == NOT_AN_ANCHOR) && 
        (coordinates.x == NULL_COORDINATE)) {
        localizeThisNode();
      }
      if (!harvestingData) {
        //gradientDemo();
        lineDemo(); 
      }
    }
    else {
    // doesn't have all pings.
    }
  } 
}

/* 
 * External Interrupt routines /INT0 and /INT1 and PCA timer overflow
 * are used to make the time of flight measurement.  The
 * Flash detector circuit triggers /INT0, which starts 
 * PCA timer and enables the sonar interrupt, /INT1.  The sonar
 * ping triggers /INT1 which makes the time of flight
 * measurement and resets all interrupt routines for the next
 * flash and ping.  If the PCA timer overflows while waiting for a
 * sonar ping, PCA interrupt is called, which simply
 * resets the system without making a time of flight measurement.
 *
 * The PCA timer runs at SYSCLOCK/12  (SYSCLOCK is 22.1184MHz).  
 * The speed of sound in air is 346.65m/s at 25 degrees C.
 * Thus, the conversion from ticks of the sysclock to meters
 * is 5317 clock ticks per meter.
 */

/*
 * External interrupt 0 is triggered by a camera flash, which signals
 * the beginning of the TOF measurement.
 * 
 * This routine does three things:
 * 1. Enable PCA timer match ISR (which is used as a timout timer)
 * 2. Enable sonar (external interrupt 1) ISR
 * 3. Disable light (external interrupt 0) ISR
 */
void external0Interrupt(void) interrupt 0 {
  /*
   * Record the current time and enable the hardware timer.  
   * This is the start time we will later use to measure the 
   * time of flight measurement.
   */
  PCA0CPL3 = PCA0L - 1;
  PCA0CPH3 = PCA0H;
  clearPCA3InterruptFlag();
  enablePCA3Interrupts();
  
  /* 
   * Enable the sonar interrupt and disable the 
   * light interrupt. It is *very* improtant that the sonar
   * interrupt flag is cleared before the interrupt is enabled, since
   * the sonar interrupt is triggered multiple times by each sonar ping,
   * and only the first interrupt is handled (and cleared) by the sonar
   * ISR.
   */
  clearSonarInterruptFlag();  // This line is important! 
  enableSonarInterrupt();   
  disableLightInterrupt(); 
}

/*
 * External interrupt 1 is triggered by a sonar ping, which signals the end of
 * the TOF range measurement.
 * 
 * This routine does the following:
 * 1. Disable PCA timer match ISR
 * 2. Disable Sonar ISR
 * 3. Calculate the time of flight for this ping
 * 4. Start the leader election algorithm by sending a leader election packet.
 * 5. Enable the camera flash ISR.
 */
void external1Interrupt(void) interrupt 2 {
  float n;
  AddressableInt stopTime, startTime;
  unsigned int difference;	

  stopPCARunControl();
  stopTime.Char[0] = PCA0L;
  stopTime.Char[1] = PCA0H;
  startPCARunControl();

  startTime.Char[0] = PCA0CPL3;
  startTime.Char[1] = PCA0CPH3;
  
  disablePCA3Interrupts();
  disableSonarInterrupt();
  
  if (stopTime.Int > startTime.Int) {
    difference = stopTime.Int - startTime.Int;
  }
  else {
    difference = stopTime.Int + (0xFFFF - startTime.Int);
  }

  /*
   * Adjust measurement to compensate for the variation with 
   * distance of the time the sonar takes to ring up.
   */
  difference -= ((unsigned int) (RING_UP_ERROR_SLOPE*((float) difference))) + RING_UP_ERROR_INTERCEPT;

  /* 
   * Record this measurement and tranmit a TOFElectMessage.
   */
  if (currentPing < NUMBER_OF_PINGS_TO_STORE) {
    pingInformation[currentPing].localDistance = difference;
    pingInformation[currentPing].minimumKnownDistance = difference;
    sendTOFElectMessage(currentPing, 
                        pingInformation[currentPing].localDistance, 
                        NeighborhoodID);
    currentPing++;
    
    /* 
     * Start the election process.  We initially assume that we are elected.
     */
    electionStartTime = getTime();
    elected = 1;
  }

   /*
   * Color the LED according to the distance of the
   * pinger.
   */

  // Place limits on the swing of the difference
  difference = ((difference > MIN_TOF_DISTANCE) 
		? difference : MIN_TOF_DISTANCE);
  difference = ((difference < MAX_TOF_DISTANCE) 
		? difference : MAX_TOF_DISTANCE);

  n = (float)(difference - MIN_TOF_DISTANCE) 
    / (float)(MAX_TOF_DISTANCE - MIN_TOF_DISTANCE) * 255.0;
  
  setRGBLEDIntensity(RGB_BLUE, 0xFF);
  setRGBLEDIntensity(RGB_RED, (unsigned char)(n));
  setRGBLEDIntensity(RGB_GREEN, 0xFF - (unsigned char)(n));
  
  clearLightInterruptFlag();
  enableLightInterrupt();  
}

/* The PCA interrupt serves as a timeout.
 * If we ever get to the PCA timer interrupt, 
 * then our sonar ping was more than 12 meters away!
 */
void pcaInterrupt(void) interrupt 9 {
  disableSonarInterrupt();
  disablePCA3Interrupts();
  clearPCA3InterruptFlag();
  
  // Indicate a timeout on the RGB LED
  setLEDState( LEDSTATE_OFF, 0xFF );

  /*
   * We have missed this sonar ping.  Move on to the next ping.
   */
  currentPing++;

  clearLightInterruptFlag();
  enableLightInterrupt();
}
