/*
 * Josh Lifton and Michael Broxton and Daniel Alejandro Benitez
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/***********************************************************************
 * TOFOutput.c
 * target: PushpinTOF
 *
 * Last modified 07JUL2004 by erlkonig.
 *
 * Functions pertaining to actuation of the RGB LED on the Pushpin TOF Module.
 * 
 * The function updateLEDState() is called repeatedly from the TOFRunLoop() 
 * routine so that it can update the state of the LED.  The updateLEDState() 
 * function will change the LED state at frequency of TOFLED_UPDATE_RATE.  
 * The exact pattern of this update depends on the ledState variable.  
 * 
 * A Low Frequency Oscillator (LFO) is incremented at the frequency of 
 * TOFLED_UPDATE_RATE. The frequency of the LFO envelope is determined by 
 * the maximum amplitude value, LFO_Max.  The value LFO_Amp represents the 
 * amplitude of the LFO at any given time and may be employed in the 
 * modification of any output values, such as the intensity of a RGB color.
 * 
 * Functionality for a computationally forgiving way by which to determine 
 * the square or integer square root of a number is included in the form of 
 * two functions, TableSquare(value) and TableSqrt(value).  These functions 
 * walk the array TableOfSquares in order to determine their return values.
 *
 * All solid LED RGB color values have been named according to the HTML 3.2 
 * color standard.
 *
 * -----Room for improvement:
 * --------scaling functions: LFO_AmpToHex() and helperColorByVolume().
 * --------squareroot table could be replaced with something smaller?
 * --------better ways to store preset colors?
 ************************************************************************/
#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>
#include <TOF.h>

#define CLOCK_SPEED 1900000			/* Number of processor clock ticks per second (1.9 MHz) */
#define TOFLED_UPDATE_RATE 61440	/* Defines desired number of clock ticks between updates (30 Hz) */

/*
 * For the following array, the 'n'th element contains the value 'n'^2.
 * This table may be used to lookup both the square or the integer square root, rounded down, of a number.
 * Use of this table has the advantage of saving the processor a great deal of computation.
 *
 * Two functions in particular, TableSquare(value) and TableSqrt(value), will call on TableOfSquares.
 */
const unsigned int TableOfSquares[102] = 
	{ 0, 1, 4, 9, 16, 25, 36, 49, 64, 81, 100, 121, 144, 169, 196, 225, 256, 289, 324, 361, 400, 
	441, 484, 529, 576, 625, 676, 729, 784, 841, 900, 961, 1024, 1089, 1156, 1225, 1296, 1369, 1444, 
	1521, 1600, 1681, 1764, 1849, 1936, 2025, 2116, 2209, 2304, 2401, 2500, 2601, 2704, 2809, 2916, 
	3025, 3136, 3249, 3364, 3481, 3600, 3721, 3844, 3969, 4096, 4225, 4356, 4489, 4624, 4761, 4900, 
	5041, 5184, 5329, 5476, 5625, 5776, 5929, 6084, 6241, 6400, 6561, 6724, 6889, 7056, 7225, 7396, 
	7569, 7744, 7921, 8100, 8281, 8464, 8649, 8836, 9025, 9216, 9409, 9604, 9801, 10000, 10201 };

unsigned char ledState;                  // The current state of the RGB LED.  See updateLEDState()...
unsigned char sweepType;		 // The desired color sweep mode.  Possible types can be found in the SweepTypes enum in TOF.h
unsigned char cycleValue = 0;

unsigned long timeOfLastUpdate;
unsigned char pulseRedValue,pulseGreenValue,pulseBlueValue;

// These arrays define the RGB values for preset colors, named according to HTML 3.2.
// They will referenced by setRGBLEDPreset().
const unsigned char OFF[3] = { 0xFF, 0xFF, 0xFF }; 
const unsigned char WHITE[3] = { 0x00, 0x00, 0x00 };
const unsigned char RED[3] = { 0x00, 0xFF, 0xFF };
const unsigned char LIME[3] = { 0xFF, 0x00, 0xFF };
const unsigned char BLUE[3] = { 0xFF, 0xFF, 0x00 };
const unsigned char AQUA[3] = { 0xFF, 0x00, 0x00 };
const unsigned char FUCHSIA[3] = { 0x00, 0xFF, 0x00 };
const unsigned char YELLOW[3] = { 0x00, 0x00, 0xFF };
const unsigned char MAROON[3] = { 0x7F, 0xFF, 0xFF };
const unsigned char GREEN[3] = { 0xFF, 0x7F, 0xFF };
const unsigned char NAVY[3] = { 0xFF, 0xFF, 0x7F };
const unsigned char TEAL[3] = { 0xFF, 0x7F, 0x7F };
const unsigned char PURPLE[3] = { 0x7F, 0xFF, 0x7F };
const unsigned char OLIVE[3] = { 0x7F, 0x7F, 0xFF };
const unsigned char GRAY[3] = { 0x7F, 0x7F, 0x7F };
const unsigned char SILVER[3] = { 0x3F, 0x3F, 0x3F };
// Happy, special, pretty colors
const unsigned char LAWNGREEN[3] = { 0x83, 0x03, 0xFF };
const unsigned char GOLD[3] = { 0x00, 0x28, 0xFF };
const unsigned char DEEPSKYBLUE[3] = { 0xFF, 0x40, 0x00 };
const unsigned char DARKORANGE[3] = { 0x00, 0x73, 0xFF };
const unsigned char DARKVIOLET[3] = { 0x6B, 0xFF, 0x2C };
const unsigned char CUSTOMRED1[3] = { 0x00, 0xFF, 0x55 };

// Variable arrays that hold the values of the colors currently in use.
unsigned char color1[3] = { 0xFF, 0xFF, 0xFF };
unsigned char color2[3] = { 0xFF, 0xFF, 0xFF };

// Control the overall, master intensity of LED output.  0xFF is full "volume" while 0x00 is effectively "off".
unsigned char colorVolume = 0xFF;

/*
 * The following variables establish and control the LFO.
 * ------------------------------------------------------
 * 
 * LFO_Amp will be scaled with LFO_Max at each implementation
 * and is incremented or decrimented by '1' during each update cycle.
 */
signed int LFO_Slope;					// '1' indicates positive slope, '-1' negative slope
unsigned int LFO_Amp;					// Current amplitude of LFO
unsigned int LFO_Max;					// Max. amp. of LFO, calculated w/ updateLFOFreq( LFO_Freq ).

/* 
 * Function Prototypes 
 */
void setColorOFF( unsigned char offColor[] );
void scaleColorByVolume( unsigned char targetColor[], unsigned char volumedColor[] );
unsigned char helperColorByVolume( unsigned char targetValue );
void updateRGBLFOPresetSweep( unsigned char bottomPreset[], unsigned char topPreset[], unsigned char sweepType );
unsigned char helperRGBLFOPresetSweep( unsigned char bottomValue, unsigned char topValue, unsigned char sweepType );
void updateRGBSparkle( unsigned char presetOne[], unsigned char presetTwo[] );


/*
 * Initialize the TOF Module output
 */
void initializeTOFOutput(void) 
{
  LFO_Slope = -1;					  // Initialize LFO, negative slope
  LFO_Amp = 0;						  // Initialize LFO, zero amplitude
  setLFOFreq( 1000 );					  // Initialize LFO with desired default frequency, 1 Hz	
  
  setLEDState( LEDSTATE_OFF, 0xFF );			  // Defines desired initial mode of LED operation
}

/*
 * Calls on TableOfSquares to lookup the integer square root, rounded down, of the argument.
 * Only valid for 0 < 'value' < 10200.
 */
unsigned int TableSqrt( unsigned int value )
{
  unsigned int n;
  
  if ( value > 10200 ) {			 // Make sure 'value' <=10200
    return 0;
  }
  
  for ( n = 0 ; n <= 100 ; n += 1 ) {  	// Walk down the lookup table until the square root is found.
    if ( value < TableOfSquares[ ( n + 1 ) ] ) {
      return n; 
    }
  }

  return 0;
}

/*
 * Scales LFO_Amp value to the appropriate hexidecimal value, 
 * linearly if scaleType = 0, parabolically otherwise, and returns it as an integer.
 * LFO_Amp values range from 0 to LFO_Max.
 * Hexidecimal values range from 0x00 to 'hexMax'.
 */
unsigned char LFO_AmpToHex( unsigned char hexMax, unsigned char scaleType )
{
  if ( scaleType == 0 )
    return ( ( LFO_Amp * hexMax ) / LFO_Max );
  else
    return ( ( TableSqrt( ( 10000 / LFO_Max ) * LFO_Amp ) * hexMax ) / 100 );
}

/*
 * Update the LFO amplitude and slope values as appropriate.
 * 
 * Make sure that LFO_Amp values do not go below 0 or above LFO_Max.
 */
void updateLFO( void )
{
  if ( LFO_Amp <= 0 ) {														
    LFO_Amp = 0;
    LFO_Slope = 1;
  }
  else if ( LFO_Amp >= LFO_Max ) {
    LFO_Amp = LFO_Max;
    LFO_Slope = -1;
  }
  
  LFO_Amp += LFO_Slope;			// Increments LFO_Amp according to LFO_Slope
}

/*
 * Take the desired LFO frequency in mHz as an argument
 * and calculate the LFO max. amp. value appropriate to TOF update rate.
 * 
 * Note that LFO_Freq is in units of mHz, not MHz!
 */
void setLFOFreq( unsigned int LFO_Freq ) {
  LFO_Max = ( ( CLOCK_SPEED * 1000 ) / 
	      ( LFO_Freq * TOFLED_UPDATE_RATE ) );
}

/*
 * Accept RGB values and pass them to setRGBLEDIntensity().
 */
void setRGBLEDDisplay( unsigned char red, unsigned char green, unsigned char blue )
{
  setRGBLEDIntensity( RGB_RED, red );
  setRGBLEDIntensity( RGB_GREEN, green );
  setRGBLEDIntensity( RGB_BLUE, blue );
}

/*
 * Accepts names of preset HTML 3.2 colors and passes the appropriate RGB values to setRGBLEDDisplay().
 */
void setRGBLEDPreset( unsigned char preset[] )
{
  setRGBLEDDisplay( preset[0], preset[1], preset[2] );
}

// Sets the master LED color intensity level in 'colorVolume'.
void setColorVolume( unsigned char volumeValue )
{
  setLEDState( ledState, volumeValue );
}

/*
 * Sets a color array to all 0xFF.
 */
void setColorOFF( unsigned char offColor[] )
{
  unsigned char n;

  for ( n = 0; n <= 2; n += 1)
    offColor[n] = 0xFF;
}

/*
 * Scale intensities of passed color according to 'colorVolume'.
 */
void scaleColorByVolume( unsigned char targetColor[], unsigned char volumedColor[] )
{
  unsigned char n;
  
  for ( n = 0; n <= 2; n += 1)
    volumedColor[n] = helperColorByVolume( targetColor[n] );
}

/*
 * Helps scaleColorByVolume().
 */
unsigned char helperColorByVolume( unsigned char targetValue )
{
  return ( 0xFF - ( ( unsigned int )( ( 0xFF - targetValue ) * colorVolume ) / 0xFF ) );
}

/*
 * Sweep RGB values between those of the presets passed to it at the LFO frequency.
 * Call helperRGBLFOPresetSweep() to calculate the appropriate values.
 */
void updateRGBLFOPresetSweep( unsigned char bottomPreset[], unsigned char topPreset[], unsigned char sweepType )
{
  pulseRedValue = helperRGBLFOPresetSweep( bottomPreset[0], topPreset[0], sweepType );
  pulseGreenValue = helperRGBLFOPresetSweep( bottomPreset[1], topPreset[1], sweepType );
  pulseBlueValue = helperRGBLFOPresetSweep( bottomPreset[2], topPreset[2], sweepType );
  
  setRGBLEDDisplay( pulseRedValue, pulseGreenValue, pulseBlueValue );	
}

/*
 * Take care of scaling the values passed by updateRGBLFOPresetSweep().
 */
unsigned char helperRGBLFOPresetSweep( unsigned char bottomValue, unsigned char topValue, unsigned char sweepType )
{
  if ( bottomValue == topValue )
    return bottomValue;
  else if ( bottomValue < topValue )
    return ( bottomValue + LFO_AmpToHex( ( topValue - bottomValue ), sweepType ) );
  else
    return ( bottomValue - LFO_AmpToHex( ( bottomValue - topValue ), sweepType ) ); 
}

/*
 * Sparkle the two presets passed as arguments at the LFO frequency.
 */
void updateRGBSparkle( unsigned char presetOne[], unsigned char presetTwo[] )
{
  if ( LFO_Slope == 1 )
    setRGBLEDPreset( presetOne );
  else setRGBLEDPreset( presetTwo );
}

void setLEDState( unsigned char stateLED, unsigned char volumeLED )
{
  ledState = stateLED;
  colorVolume = volumeLED;
  
  switch ( ledState ) 
    {
    case LEDSTATE_WHITE :
      scaleColorByVolume( WHITE, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_RED :
      scaleColorByVolume( RED, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_LIME :
      scaleColorByVolume( LIME, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_BLUE :
      scaleColorByVolume( BLUE, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_AQUA :
      scaleColorByVolume( AQUA, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_FUCHSIA :
      scaleColorByVolume( FUCHSIA, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_YELLOW :
      scaleColorByVolume( YELLOW, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_MAROON :
      scaleColorByVolume( MAROON, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_GREEN :
      scaleColorByVolume( GREEN, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_NAVY :
      scaleColorByVolume( NAVY, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_TEAL :
      scaleColorByVolume( TEAL, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_PURPLE :
      scaleColorByVolume( PURPLE, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_OLIVE :
      scaleColorByVolume( OLIVE, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_GRAY :
      scaleColorByVolume( GRAY, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_SILVER :
      scaleColorByVolume( SILVER, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_DARKORANGE :
      scaleColorByVolume( DARKORANGE, color1 );
      setColorOFF( color2 );
      sweepType = SOLID;
      break;
    case LEDSTATE_OFF :
      setColorOFF( color1 );
      setColorOFF( color2 );
      break;
    /* 
     * The following pulse the intensity of a single RGB value, either linearly or parabolically.
     */
    case LEDSTATE_REDLINEAR :							
      setColorOFF( color1 );
      scaleColorByVolume( RED, color2 );
      sweepType = LINEAR;
      setLFOFreq( 1000 );							
      break;
    case LEDSTATE_REDPARA :								
      setColorOFF( color1 );
      scaleColorByVolume( RED, color2 );
      sweepType = PARABOLIC;
      setLFOFreq( 1000 );							
      break;
    case LEDSTATE_GREENLINEAR :							
      setColorOFF( color1 );
      scaleColorByVolume( GREEN, color2 );
      sweepType = LINEAR;
      setLFOFreq( 1000 );							
      break;
    case LEDSTATE_GREENPARA :							
      setColorOFF( color1 );
      scaleColorByVolume( GREEN, color2 );
      sweepType = PARABOLIC;
      setLFOFreq( 1000 );							
      break;
    case LEDSTATE_BLUELINEAR :							
      setColorOFF( color1 );
      scaleColorByVolume( BLUE, color2 );
      sweepType = LINEAR;
      setLFOFreq( 1000 );							
      break;
    case LEDSTATE_BLUEPARA :							
      setColorOFF( color1 );
      scaleColorByVolume( BLUE, color2 );
      sweepType = PARABOLIC;
      setLFOFreq( 1000 );							
      break;
    case LEDSTATE_PULSE4 :			// Strobe like an airplane running light.
      setColorOFF( color1 );
      scaleColorByVolume( BLUE, color2 );
      sweepType = STROBE;
      setLFOFreq( 1200 );
      break;
    case LEDSTATE_TOFCHILLIN :
      scaleColorByVolume( BLUE, color1 );
      scaleColorByVolume( NAVY, color2 );
      sweepType = SQUARE;
      setLFOFreq( 10000 );
      break;
    case LEDSTATE_TOFEVERYONE :
      scaleColorByVolume( BLUE, color1 );
      scaleColorByVolume( NAVY, color2 );
      sweepType = LINEAR;
      setLFOFreq( 1000 );
      break;
    case LEDSTATE_TOFANCHOR1 :
      scaleColorByVolume( YELLOW, color1 );
      scaleColorByVolume( RED, color2 );
      sweepType = PARABOLIC;
      setLFOFreq( 5000 );
      break;
    case LEDSTATE_TOFANCHOR2 :
      scaleColorByVolume( YELLOW, color1 );
      scaleColorByVolume( GREEN, color2 );
      sweepType = PARABOLIC;
      setLFOFreq( 5000 );
      break;
    case LEDSTATE_TOFANCHOR3 :
      scaleColorByVolume( YELLOW, color1 );
      scaleColorByVolume( BLUE, color2 );
      sweepType = PARABOLIC;
      setLFOFreq( 5000 );
      break;
    case LEDSTATE_TOFBOUND1 :
      scaleColorByVolume( RED, color1 );
      scaleColorByVolume( CUSTOMRED1, color2 );
      sweepType = LINEAR;
      setLFOFreq( 500 );
      break;
    case LEDSTATE_TOFBOUND2 :
      scaleColorByVolume( GREEN, color1 );
      scaleColorByVolume( BLUE, color2 );
      sweepType = LINEAR;
      setLFOFreq( 500 );
      break;
    case LEDSTATE_TEST :  // Meme: for testing.
      setLFOFreq( 1000 );
      //		updateRGBLFOPresetSweep( WHITE, GREEN, 1 );
      //		updateRGBLFOPresetSweep( GOLD, DARKORANGE, 0 );
      //		updateRGBSparkle( GOLD, DARKORANGE );
      //		updateRGBSparkle( OLIVE, maximizeRGBColor( OLIVE ) );
      break;
    }
}


/*
 * This function set the LED to a color given by the input from the Debugger GUI.
 * Instead of using only preset color, the user can input any color it wants.
 */

void setInputLEDState(unsigned char color[]) {
	scaleColorByVolume( color, color1);
	setColorOFF( color2 );
	sweepType = SOLID;
}

/*
 * This is the main update routine for the actuation on the TOF module.  This
 * routine is called regularly by TOFRunLoop() to update the state of the output
 * (in this case, the output consists of only the RGB LED).  This routine is responsible
 * for creating a consistent update rate (currently 30 Hz) and calling the appropriate 
 * LED update function based on the current value of the variable ledState.
 */
void updateLEDState( void ) 
{
  /*
   * This function runs on a cycle of TOFLED_UPDATE_RATE cycles per second.
   * If less than one update cycle has passed, return without updating the LED.
   */
  if ( getTime() - timeOfLastUpdate < TOFLED_UPDATE_RATE ) 
    {
      return;
    }
  
  timeOfLastUpdate = getTime();
  
  switch ( sweepType )
    {
    case SOLID :
      setRGBLEDPreset( color1 );
      break;
    case LINEAR :
      updateRGBLFOPresetSweep( color1, color2, 0);
      break;
    case PARABOLIC :
      updateRGBLFOPresetSweep( color1, color2, 1);
      break;
    case SQUARE :
      updateRGBSparkle( color1, color2 );
      break;
    case STROBE :
      if ( LFO_Slope == 1 )
	updateRGBLFOPresetSweep( color1, color2, 1);
      else clearRGBLEDDisplay();
      break;
    }

  updateLFO();		// Runs function that updates LFO amp. and slope values.
}


